/*
 * @(#)Bootloader.java	0.1 02/09/26 SMI
 *
 * Copyright (c) 2002, 2003 Martin Dubuc All Rights Reserved.
 *
 * Martin Dubuc grants you ("Licensee") a non-exclusive, royalty free, license
 * to use, modify and redistribute this software in source and binary
 * code form, provided that i) this copyright notice and license appear
 * on all copies of the software; and ii) Licensee does not utilize the
 * software in a manner which is disparaging to Martin Dubuc.
 *
 * This software is provided "AS IS," without a warranty of any kind.
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. MARTIN DUBUC
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THE
 * SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL MARTIN DUBUC OR ITS LICENSORS
 * BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES,
 * HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING
 * OUT OF THE USE OF OR INABILITY TO USE SOFTWARE, EVEN IF MARTIN DUBUC
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control
 * of aircraft, air traffic, aircraft navigation or aircraft
 * communications; or in the design, construction, operation or
 * maintenance of any nuclear facility. Licensee represents and
 * warrants that it will not use or redistribute the Software for such
 * purposes.
 */

import javax.comm.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.URL;
import java.util.Properties;
import java.util.Enumeration;
import javax.swing.*;

/**
 Main file for Bootloader program. This file contains the GUI framework that
 the program runs in.
*/

public class Bootloader extends JFrame implements ActionListener, ItemListener {
  final static long serialVersionUID = 0;

  final static String PRODUCT_NAME = "Jolt PIC 18F Bootloader";
  
  final private static int HEIGHT = 470;
  final private static int WIDTH  = 410;

  // Hex file handling definitions
  final private int CODE_DATA = 0;
  final private int CODE_EOF = 1;
  final private int CODE_ADDR_HI = 4;
  final private int ADDR_HI_CONFIG = 48;
  final private int ADDR_HI_EE_DATA = 240;
  final private int DATA_COUNT = 16;

  // Command termination reason definitions
  public static final int COMMAND_TERM_REASON_SUCCESS = 0;
  public static final int COMMAND_TERM_REASON_PORT_OPEN_FAILURE = 1;
  public static final int COMMAND_TERM_REASON_PORT_BUSY = 2;
  public static final int COMMAND_TERM_REASON_RESOURCE_NOT_AVAILABLE = 3;
  public static final int COMMAND_TERM_REASON_TIMEOUT = 4;
  public static final int COMMAND_TERM_REASON_USER = 5;
  public static final int COMMAND_TERM_REASON_VERIFY_FAILURE = 6;
  public static final int COMMAND_TERM_REASON_UNEXPECTED_MESSAGE = 7;

  // Packet field offset definition
  private final int COMMAND  = 0;
  private final int DATALEN  = 1;
  private final int ADDRL    = 2;
  private final int ADDRH    = 3;
  private final int ADDRU    = 4;
  private final int DATA     = 5;
  private final int VERSIONL = 2;
  private final int VERSIONH = 3;

  // Command definitions
  private final int CommandNull               = -1;
  private final int CommandReadVersion        = 0x00;
  private final int CommandReadProgramMemory  = 0x01; 
  private final int CommandWriteProgramMemory = 0x02;
  private final int CommandEraseProgramMemory = 0x03;
  private final int CommandReadEeData         = 0x04;
  private final int CommandWriteEeData        = 0x05;
  private final int CommandReadConfig         = 0x06;
  private final int CommandWriteConfig        = 0x07;
  private final int CommandReset              = 0x08;

  private final int EXIT_NORMAL = 0;
  private final int EXIT_ERROR = 1;
  
  private JMenuBar mb;
  private JMenu fileMenu;
  private JMenu editMenu;
  private JMenu commandMenu;
  private JMenu settingsMenu;
  private JMenu helpMenu;
  private JMenuItem loadItem;
  private JMenuItem saveItem;
  private JMenuItem saveAsItem;
  private JMenuItem exitItem;
  private JMenuItem configItem;
  private JMenuItem eeDataItem;
  private JMenuItem readItem;
  private JMenuItem programItem;
  private JMenuItem writeConfigItem;
  private JMenuItem writeDataItem;
  private JMenuItem verifyItem;
  private JMenuItem runItem;
  private JMenu bootModeMenu;
  private JRadioButtonMenuItem delayedItem;
  private JRadioButtonMenuItem userDefinedItem;
  private JCheckBoxMenuItem programAutoloadItem;
  private JCheckBoxMenuItem programResetItem;
  private JCheckBoxMenuItem programConfigItem;
  private JCheckBoxMenuItem programEeDataItem;
  private JMenu languageMenu;
  private JRadioButtonMenuItem englishItem;
  private JRadioButtonMenuItem frenchItem;
  private JMenu logMenu;
  private JRadioButtonMenuItem debugItem;
  private JRadioButtonMenuItem infoItem;
  private JRadioButtonMenuItem errorItem;
  private JRadioButtonMenuItem noneItem;
  private JMenuItem helpItem;
  private JMenuItem aboutItem;

  private JLabel firmwareVersionHeader;
  private JTextField firmwareVersionField;
  private JLabel deviceIdHeader;
  private Choice deviceIdChoice;
  private JPanel devicePanel;

  private JButton sendButton;

  private JTextField sendLine;

  private JPanel messagePanel;
  private JScrollPane memoryAreaScrollPane;
  private JScrollBar verticalScrollBar;
  private JTextArea memoryArea;
  private JTextArea messageAreaIn;

  private ConfigurationPanel configurationPanel;
  private BootloaderParameters parameters;
  private SerialConnection connection;

  private Log log;

  private String cliFilename;
  
  private boolean cliVerbose = false;
  
  private ProgressDialog progressDialog;
  private int lastProgress = 0;
  
  private Timer refreshTimer;

  private Properties props = null;

  private Cursor previousCursor;
  
  private PicDeviceTable deviceTable = null;

  private int firmwareVersion;
  private int deviceId;
    
  private PicDevice device;

  private long id;
  
  final private int WRITE_DATA_SIZE = 16;

  private Timer retryTimer;

  private byte lastCommandSent[];
  private int lastCommandSentLength;

  private int currentTransaction;
  private int currentCommand;
  private int currentOffset;
  private int numProgramBlock;
  private int lastWriteOffset;
  private int timeout;
  private int maxTimeouts;
  private int percent, lastPercent;
  private boolean verifyInd; // Indicates that bootloader is performing a verification
  private boolean eeDataInd; // Indicates that bootloader is programming EE Data only
  private boolean configInd; // Indicates that bootloader is programming Config only
  private boolean startupDelay;
  byte codeSpace[];
  byte eeData[];
  byte config[];
  private int deviceIdRaw[] = new int[2];

  /**
  Main method. Checks to see if the command line agrument is requesting
  usage informaition (-h, -help), if it is, display a usage message and
  exit, otherwise create a new <code>Bootloader</code> and set it visible.
  */
  public static void main(String[] args) {
    if ((args.length > 0) && (args[0].equals("-h") ||
        args[0].equals("-help"))) {
      System.out.println("usage: java -cp Bootloader.jar Bootloader [-r] [-s] [-b<baudrate>] [firmware.hex]");
      System.out.println("Options:");
      System.out.println("    -r  reset PIC after program");
      System.out.println("    -s  stay in boot after program");
      System.out.println("    -b  set baud rate [4800 | 9600 | 19200 | 38400 | 57600 | 115200]\n");
      System.exit(1);
    }

    setLookAndFeel();
    
    Bootloader bootloader = new Bootloader(args);
    
    // Force window refresh
    bootloader.repaint();
  } // main

  private static void setLookAndFeel() {
	String lnf = UIManager.getSystemLookAndFeelClassName();
    
	try {
	  // Set look and feel to default system look and feel
	  UIManager.setLookAndFeel(lnf);
	}
	catch(Exception e) {
	  System.out.println("Exception " + e + ": Could not set system look and feel.");
	}
  } // setLookAndFeel
  
  /**
  Create new <code>Bootloader</code> and initilizes it. Parses args to
  find configuration file. If found, initial state it set to parameters
  in configuration file.

  @param args command line arguments used when program was invoked.
  */
  public Bootloader(String[] args) {
    super(PRODUCT_NAME);
    
    id = System.currentTimeMillis();
    
    parameters = new BootloaderParameters();

    loadParameters();

    parseArguments(args);
    
    BootloaderResourceBundle.setLanguage(parameters.getLanguage());
    
    loadDeviceTable();

    messageAreaIn = new JTextArea();

    log = new Log(messageAreaIn);
    log.setSeverityLevel(parameters.getLogLevel());
   
	initContext();

    // Set up the GUI for the program
    Container contentPane = getContentPane();

    addWindowListener(new CloseHandler(this));

    mb = new JMenuBar();

    fileMenu = new JMenu(BootloaderResourceBundle.getString("FILE_TEXT"));
    fileMenu.setMnemonic('F');

    loadItem = new JMenuItem(BootloaderResourceBundle.getString("OPEN_HEX_FILE_TEXT") + "...", 'O');
    loadItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, 
                                                   InputEvent.CTRL_MASK));
    loadItem.addActionListener(this);
    fileMenu.add(loadItem);

    saveItem = new JMenuItem(BootloaderResourceBundle.getString("SAVE_HEX_FILE_TEXT"));
    if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  saveItem.setMnemonic('S');
      saveItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, 
                                                     InputEvent.CTRL_MASK));
    }
    saveItem.addActionListener(this);
    fileMenu.add(saveItem);

    saveAsItem = new JMenuItem(BootloaderResourceBundle.getString("SAVE_HEX_FILE_AS_TEXT") + "...");
	if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  saveAsItem.setMnemonic('A');
	}
    saveAsItem.addActionListener(this);
    fileMenu.add(saveAsItem);

    fileMenu.addSeparator();

    exitItem = new JMenuItem(BootloaderResourceBundle.getString("EXIT_TEXT"));
	if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  exitItem.setMnemonic('X');
	}
	else if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_FRENCH) {
	  exitItem.setMnemonic('Q');
	}
	
    exitItem.addActionListener(this);
    fileMenu.add(exitItem);

    addRecentFileMenu();

    mb.add(fileMenu);

    editMenu = new JMenu(BootloaderResourceBundle.getString("EDIT_TEXT"));
    editMenu.setMnemonic('E');

    configItem = new JMenuItem(BootloaderResourceBundle.getString("CONFIG_TEXT"), 'C');
    configItem.addActionListener(this);
    editMenu.add(configItem);

    eeDataItem = new JMenuItem(BootloaderResourceBundle.getString("EEPROM_DATA_TEXT"));
	if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  eeDataItem.setMnemonic('E');
	}
	else if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_FRENCH) {
	  eeDataItem.setMnemonic('D');
	}
    eeDataItem.addActionListener(this);
    editMenu.add(eeDataItem);

    mb.add(editMenu);

    commandMenu = new JMenu(BootloaderResourceBundle.getString("COMMAND_TEXT"));
    commandMenu.setMnemonic('C');

    readItem = new JMenuItem(BootloaderResourceBundle.getString("READ_ALL_TEXT"));
    if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  readItem.setMnemonic('R');
	  readItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_R, 
													 InputEvent.CTRL_MASK));
	}
    else if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_FRENCH) {
	  readItem.setMnemonic('L');
	  readItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_L, 
													 InputEvent.CTRL_MASK));
	}
    readItem.addActionListener(this);
    commandMenu.add(readItem);

    programItem = new JMenuItem(BootloaderResourceBundle.getString("PROGRAM_TEXT"), 'P');
	programItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F4, 0));
    programItem.addActionListener(this);
    commandMenu.add(programItem);

    writeConfigItem = new JMenuItem(BootloaderResourceBundle.getString("PROGRAM_CONFIG_TEXT"), 'C');
    writeConfigItem.addActionListener(this);
    commandMenu.add(writeConfigItem);

    writeDataItem = new JMenuItem(BootloaderResourceBundle.getString("PROGRAM_EEPROM_DATA_TEXT"));
	if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  writeDataItem.setMnemonic('E');
    }
    writeDataItem.addActionListener(this);
    commandMenu.add(writeDataItem);

    verifyItem = new JMenuItem(BootloaderResourceBundle.getString("VERIFY_TEXT"), 'V');
    verifyItem.addActionListener(this);
    commandMenu.add(verifyItem);

    commandMenu.addSeparator();

    runItem = new JMenuItem(BootloaderResourceBundle.getString("RUN_TEXT"));
	if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  runItem.setMnemonic('R');
	}
	else if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_FRENCH) {
	  runItem.setMnemonic('E');
	}
    runItem.addActionListener(this);
    commandMenu.add(runItem);

    mb.add(commandMenu);

    settingsMenu = new JMenu(BootloaderResourceBundle.getString("SETTINGS_TEXT"));
    if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH)
      settingsMenu.setMnemonic('S');
    else if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_FRENCH)
	  settingsMenu.setMnemonic('P');

    bootModeMenu = new JMenu(BootloaderResourceBundle.getString("BOOT_MODE_TEXT"));
    ButtonGroup bootModeGroup = new ButtonGroup();
    delayedItem = new JRadioButtonMenuItem(BootloaderResourceBundle.getString("DELAYED_TEXT"));
    delayedItem.addActionListener(this);
    if (parameters.getBootMode() == BootloaderParameters.BOOT_MODE_DELAYED)
      delayedItem.setSelected(true);
    userDefinedItem = new JRadioButtonMenuItem(BootloaderResourceBundle.getString("USER_DEFINED_TEXT"));
    userDefinedItem.addActionListener(this);
    if (parameters.getBootMode() == BootloaderParameters.BOOT_MODE_USER)
      userDefinedItem.setSelected(true);
    bootModeGroup.add(delayedItem);
    bootModeGroup.add(userDefinedItem);
    bootModeMenu.add(delayedItem);
    bootModeMenu.add(userDefinedItem);
    settingsMenu.add(bootModeMenu);

    programAutoloadItem = new JCheckBoxMenuItem(BootloaderResourceBundle.getString("RELOAD_BEFORE_PROGRAM_TEXT"));
    programAutoloadItem.addActionListener(this);
    programAutoloadItem.setState(parameters.getProgramAutoload());
    settingsMenu.add(programAutoloadItem);

    programResetItem = new JCheckBoxMenuItem(BootloaderResourceBundle.getString("RESET_AFTER_PROGRAM_TEXT"));
    programResetItem.addActionListener(this);
    programResetItem.setState(parameters.getProgramReset());
    if (parameters.getBootMode() == BootloaderParameters.BOOT_MODE_USER) {
	  // Disable reset after program menu item
	  programResetItem.setEnabled(false);
    }
	settingsMenu.add(programResetItem);

    programConfigItem = new JCheckBoxMenuItem(BootloaderResourceBundle.getString("ENABLE_PROGRAM_CONFIG_TEXT"));
    programConfigItem.addActionListener(this);
    programConfigItem.setState(parameters.getProgramConfig());
    settingsMenu.add(programConfigItem);

    programEeDataItem = new JCheckBoxMenuItem(BootloaderResourceBundle.getString("ENABLE_PROGRAM_EEPROM_DATA_TEXT"));
    programEeDataItem.addActionListener(this);
    programEeDataItem.setState(parameters.getProgramEeData());
    settingsMenu.add(programEeDataItem);

	languageMenu = new JMenu(BootloaderResourceBundle.getString("LANGUAGE_TEXT"));
	ButtonGroup languageGroup = new ButtonGroup();
	englishItem = new JRadioButtonMenuItem(BootloaderResourceBundle.getString("ENGLISH_TEXT"));
	englishItem.addActionListener(this);
	if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH)
	  englishItem.setSelected(true);
	frenchItem = new JRadioButtonMenuItem(BootloaderResourceBundle.getString("FRENCH_TEXT"));
	frenchItem.addActionListener(this);
	if (parameters.getBootMode() == BootloaderParameters.LANGUAGE_FRENCH)
	  frenchItem.setSelected(true);
	languageGroup.add(englishItem);
	languageGroup.add(frenchItem);
	languageMenu.add(englishItem);
	languageMenu.add(frenchItem);
	settingsMenu.add(languageMenu);

    logMenu = new JMenu("Log");
    ButtonGroup group = new ButtonGroup();
    debugItem = new JRadioButtonMenuItem("Debug");
    debugItem.addActionListener(this);
    if (parameters.getLogLevel() == Log.DEBUG)
      debugItem.setSelected(true);
    infoItem = new JRadioButtonMenuItem("Info");
    infoItem.addActionListener(this);
    if (parameters.getLogLevel() == Log.INFO)
      infoItem.setSelected(true);
    errorItem = new JRadioButtonMenuItem("Error");
    errorItem.addActionListener(this);
    if (parameters.getLogLevel() == Log.ERROR)
      errorItem.setSelected(true);
    noneItem = new JRadioButtonMenuItem("None");
    noneItem.addActionListener(this);
    if (parameters.getLogLevel() == Log.NONE)
      noneItem.setSelected(true);
    group.add(debugItem);
    group.add(infoItem);
    group.add(errorItem);
    group.add(noneItem);
    logMenu.add(debugItem);
    logMenu.add(infoItem);
    logMenu.add(errorItem);
    logMenu.add(noneItem);

    if (parameters.getLogLevel() != Log.NONE) {
      settingsMenu.addSeparator();

      settingsMenu.add(logMenu);
    }

    mb.add(settingsMenu);

    helpMenu = new JMenu(BootloaderResourceBundle.getString("HELP_TEXT"));
    helpMenu.setMnemonic('H');

    helpItem = new JMenuItem(BootloaderResourceBundle.getString("CONTENTS_TEXT"));
	if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_ENGLISH) {
	  helpItem.setMnemonic('C');
	}
	else if (parameters.getLanguage() == BootloaderParameters.LANGUAGE_FRENCH) {
	  helpItem.setMnemonic('T');
	}
    helpItem.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0));
    helpItem.addActionListener(this);
    helpMenu.add(helpItem);

    helpMenu.addSeparator();

    aboutItem = new JMenuItem(BootloaderResourceBundle.getString("ABOUT_TEXT") + " " + PRODUCT_NAME + "...", 'A');
    aboutItem.addActionListener(this);
    helpMenu.add(aboutItem);

    mb.add(helpMenu);

    setJMenuBar(mb);

    devicePanel = new JPanel();
    firmwareVersionHeader = new JLabel(BootloaderResourceBundle.getString("FIRMWARE_VERSION_TEXT") + ":");
    devicePanel.add(firmwareVersionHeader);
    firmwareVersionField = new JTextField(BootloaderResourceBundle.getString("UNKNOWN_TEXT"), 5);
    firmwareVersionField.setEditable(false);
    devicePanel.add(firmwareVersionField);
	deviceIdHeader = new JLabel(BootloaderResourceBundle.getString("DEVICE_ID_TEXT") + ":");
    devicePanel.add(deviceIdHeader);
	deviceIdChoice = new Choice();
	deviceIdChoice.add(BootloaderResourceBundle.getString("UNKNOWN_TEXT") + "    ");
    deviceIdChoice.addItemListener(this);
    //Dimension d = deviceIdChoice.getSize();
	devicePanel.add(deviceIdChoice);

    contentPane.add(devicePanel, "North");

    messagePanel = new JPanel();

    if (parameters.getLogLevel() == Log.NONE)
      messagePanel.setLayout(new GridLayout(1, 1));
    else
      messagePanel.setLayout(new GridLayout(2, 1));

    memoryArea = new JTextArea();
 
    if (parameters.getLogLevel() == Log.NONE)
      memoryArea.setSize(WIDTH - 10, HEIGHT - 100);

    memoryAreaScrollPane = new JScrollPane(memoryArea);
    verticalScrollBar = memoryAreaScrollPane.getVerticalScrollBar();

    // Add special key sequence to navigate memory panel
    memoryAreaScrollPane.
    registerKeyboardAction(new ScrollableAction(memoryAreaScrollPane,
                                                JScrollBar.VERTICAL,
                                                ScrollableAction.UNIT, -1),
                           KeyStroke.getKeyStroke(KeyEvent.VK_UP, 0, false),
                           JComponent.WHEN_IN_FOCUSED_WINDOW);
    memoryAreaScrollPane.
    registerKeyboardAction(new ScrollableAction(memoryAreaScrollPane,
                                                JScrollBar.VERTICAL,
                                                ScrollableAction.BLOCK, -1),
                           KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_UP, 0, false),
                           JComponent.WHEN_IN_FOCUSED_WINDOW);
    memoryAreaScrollPane.
    registerKeyboardAction(new ScrollableAction(memoryAreaScrollPane,
                                                JScrollBar.VERTICAL, 
                                                ScrollableAction.AREA, -1),
                           KeyStroke.getKeyStroke(KeyEvent.VK_HOME, InputEvent.CTRL_MASK,
                                                  false),
                           JComponent.WHEN_IN_FOCUSED_WINDOW);
    memoryAreaScrollPane.
    registerKeyboardAction(new ScrollableAction(memoryAreaScrollPane,
                                                JScrollBar.VERTICAL, 
                                                ScrollableAction.UNIT, 1),
                           KeyStroke.getKeyStroke(KeyEvent.VK_DOWN, 0, false),
                           JComponent.WHEN_IN_FOCUSED_WINDOW);
    memoryAreaScrollPane.
    registerKeyboardAction(new ScrollableAction(memoryAreaScrollPane,
                                                JScrollBar.VERTICAL, 
                                                ScrollableAction.BLOCK, 1),
                           KeyStroke.getKeyStroke(KeyEvent.VK_PAGE_DOWN, 0, false),
                           JComponent.WHEN_IN_FOCUSED_WINDOW);
    memoryAreaScrollPane.
    registerKeyboardAction(new ScrollableAction(memoryAreaScrollPane,
                                                JScrollBar.VERTICAL,
                                                ScrollableAction.AREA, 1),
                           KeyStroke.getKeyStroke(KeyEvent.VK_END, InputEvent.CTRL_MASK,
                                                  false),
                           JComponent.WHEN_IN_FOCUSED_WINDOW);
    memoryArea.setEditable(false);
    memoryArea.setFont(new Font("Courier", Font.PLAIN, 12));
    messagePanel.add(memoryAreaScrollPane);

    JScrollPane messageAreaInScrollPane = new JScrollPane(messageAreaIn);
    messageAreaIn.setEditable(false);
    messageAreaIn.setFont(new Font("Courier", Font.PLAIN, 12));

    if (log.getSeverityLevel() != Log.NONE) {
      messagePanel.add(messageAreaInScrollPane);
    }

    contentPane.add(messagePanel, "Center");

    configurationPanel = new ConfigurationPanel(this);

    sendLine = new JTextField("", 30);

    sendButton = new JButton("Send");
    sendButton.addActionListener(this);
	
    JPanel southPanel = new JPanel();

    GridBagLayout gridBag = new GridBagLayout();
    GridBagConstraints cons = new GridBagConstraints();

    southPanel.setLayout(gridBag);

    cons.fill = GridBagConstraints.BOTH;
    cons.insets = new Insets(2, 2, 2, 2);
    cons.weightx = 1.0;
    cons.gridwidth = GridBagConstraints.REMAINDER;
    gridBag.setConstraints(configurationPanel, cons);
    southPanel.add(configurationPanel);

    contentPane.add(southPanel, "South");

    setConfigurationPanel();
	
    Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();

    connection = new SerialConnection(this, parameters, log);

    ActionListener timeoutAction = new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        verticalScrollBar.setValue(verticalScrollBar.getMinimum());
      }
    };

    refreshTimer = new Timer(1, timeoutAction);
    refreshTimer.setRepeats(false);

    int width = WIDTH;
    int height = HEIGHT;

    if (parameters.getProgramName().equals("")) {
      setTitle(PRODUCT_NAME);
      eraseMemory();
      refreshMemory();
    }
    else {
      setTitle(PRODUCT_NAME + " - " + parameters.getProgramName());
      reloadHexFile();
    }

    setLocation(screenSize.width / 2 - width / 2, 
                screenSize.height / 2 - height / 2);

    setSize(width, height);

	URL url = getClass().getResource("Jolt.gif");
	if (url != null) {
	  Image image = Toolkit.getDefaultToolkit().getImage(url);

	  setIconImage(image);
	}

	if (!isCli())
	  setVisible(true);

	if (cliFilename == null)
	  progressDialog = new ProgressDialog(this, "");
    
	deviceIdChoice.removeAll();
	deviceIdChoice.add(device.name);
	
	deviceIdRaw[0] = 0;
	deviceIdRaw[1] = 0;
	
	if (cliFilename == null)
      return;
	
	File f = new File(cliFilename);
	if (!loadFile(f))
	  System.exit(EXIT_ERROR);
	
    programDeviceTransaction();
  } // Bootloader

  private void parseArguments(String[] args) {
  	int i;
  	
  	for (i = 0; i < args.length; i++) {
      if (args[i].equals("-r"))
          	parameters.setProgramReset("true");
      else if (args[i].equals("-s"))
          	parameters.setProgramReset("false");
      else if (args[i].equals("-v"))
      	cliVerbose = true;
      else if (args[i].startsWith("-b")) {
      	String baudRate = args[i].substring(2);
      	parameters.setBaudRate(baudRate);
      }
      else
        cliFilename = args[i];
  	}
  } // parseArguments
  
  private boolean loadFile(File f) {
    int lastAddress;
  
    try {
      FileInputStream fis = new FileInputStream(f);
      lastAddress = getLastAddress(fis);
      fis.close();
      resizeMemory(lastAddress);
      eraseMemory();
      fis = new FileInputStream(f);
      loadHexFile(fis);
      fis.close();
    }
    catch (FileNotFoundException e1) {
      System.err.println(e1);
      return false;
    }
    catch (IOException e2) {
      System.err.println(e2);
      return false;
    }
    
    return true;
  } // loadFile
  
  public String getAppId() {
  	return "Jolt " + id;
  } // getAppId
  
  /**
   Event handler for changes in the current selection of the Choices.
   If a port is open the port can not be changed.
   If the choice is unsupported on the platform then the user will
   be notified and the settings will revert to their pre-selection
   state.
   */
  public void itemStateChanged(ItemEvent e) {
	if (e.getItemSelectable() == deviceIdChoice) {
	  PicDevice dev = deviceTable.get(deviceId);
	  if (dev != null) {
	  	// Determine if we need to switch to alternate device.
	    if (deviceIdChoice.getSelectedItem().endsWith("39") ||
	        deviceIdChoice.getSelectedItem().endsWith("90")) {
	      // Selected device is a XX39 or XX90 model.
	      parameters.setPic18fxx2Ind(false);
	      if (dev.name.endsWith("39") || dev.name.endsWith("90"))
	        device = dev;
	      else
	      	device = dev.alternateDevice;
	    }
	    else {
	      // Not a XX39 or XX90.
	      parameters.setPic18fxx2Ind(true);
	      if (dev.name.endsWith("39") || dev.name.endsWith("90"))
		    device = dev.alternateDevice;
		  else
		    device = dev;
	    }
	  }
	  refreshMemory();
	  refreshTimer.start();
	  saveParameters();
	}
  } // itemStateChanged
	
  private void initContext() {
	lastCommandSent = new byte[SerialConnection.MAX_BUFFER_LENGTH];
	lastCommandSentLength = 0;
	currentTransaction = CommandNull;
	currentCommand = CommandNull;
	currentOffset = 0;
	percent = 0;
	lastPercent = 0;
	verifyInd = false;
    startupDelay = false;
    
	codeSpace = new byte[PicDevice.MAX_PROGRAM_MEMORY_SIZE];
	eeData = new byte[PicDevice.MAX_EEDATA_SIZE];
	config = new byte[PicDevice.CONFIG_SIZE];
	
	device = new PicDevice(BootloaderResourceBundle.getString("UNKNOWN_TEXT"), 0x200, 0x7FFF, 0, 255,
			               "FOSC20,BORV,BOREN,PWRTEN,WDTPS20,WDTEN,DEBUG,LVP,STVREN,CP1,CP0,CPD,CPB,WRT1,WRT0,WRTD,WRTB,WRTC,EBTR1,EBTR0,EBTRB");

	ActionListener timeoutAction = new ActionListener() {
	  public void actionPerformed(ActionEvent evt) {
		retryCommand();
	  }
	};

	retryTimer = new Timer(parameters.getTimeoutDelay(), timeoutAction);
	retryTimer.setRepeats(false);
  } // initContext
  
  /**
  Add recent files in file menu.
  */
  public void addRecentFileMenu() {
    JMenuItem item;
    int i;
    boolean done = false;

    for ( ; !done; ) {
      item = fileMenu.getItem(3);
      if (item == exitItem) {
        // We have reached end of the recent list
        done = true;
      }

      fileMenu.remove(3);
    }

    fileMenu.addSeparator();

    for (i = 0; i < BootloaderParameters.NUM_RECENT_PROGRAM; i++) {
      int index;

      if (parameters.getRecentProgramName(i) == null)
        break;

      index = i + 1;
      item = new JMenuItem(index + " " + parameters.getRecentProgramName(i), '0' + index);
      item.addActionListener(this);
      fileMenu.add(item);
    }

    fileMenu.addSeparator();

    fileMenu.add(exitItem);
  } // addRecentFileMenu

  /**
  Sets the GUI elements on the configurationPanel.
  */
  public void setConfigurationPanel() {
    configurationPanel.setConfigurationPanel();
  } // setConfigurationPanel

  /**
  Responds to the menu items and buttons.
  */
  public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    int lastAddress;
    String filename;
    String dir;
    File f;

    // Load a hex file to memory
    if (cmd.equals(BootloaderResourceBundle.getString("OPEN_HEX_FILE_TEXT") + "...")) {
      if (parameters.getFileDialogUseSwing()) {
        JFileChooser chooser = new JFileChooser();
        BasicFileFilter filter = new BasicFileFilter();
        filter.addExtension("hex");
        filter.setDescription(BootloaderResourceBundle.getString("HEX_FILES_TEXT"));
        chooser.setFileFilter(filter);
        if (!parameters.getProgramDir().equals(""))
          chooser.setCurrentDirectory(new File(parameters.getProgramDir()));
        int returnVal = chooser.showOpenDialog(this);
        if (returnVal != JFileChooser.APPROVE_OPTION) {
          return;
        }
        filename = chooser.getSelectedFile().getName();
        dir = chooser.getCurrentDirectory().getPath();
        f = chooser.getSelectedFile();
      }
      else {
        FileDialog fd = new FileDialog(this, 
                                       BootloaderResourceBundle.getString("OPEN_HEX_FILE_TEXT"),
                                       FileDialog.LOAD);
        if (!parameters.getProgramDir().equals(""))
          fd.setDirectory(parameters.getProgramDir());
        fd.setFilenameFilter(new HexFilenameFilter());
        fd.setVisible(true);
        filename = fd.getFile();
        if (filename == null)
          return;
        dir = fd.getDirectory();
        f = new File(dir + filename);
      }

      if (!loadFile(f))
      	return;

      refreshMemory();
      refreshTimer.start();
      parameters.setProgramDir(dir);
      parameters.setProgramName(filename);
      parameters.setRecentProgram(dir, filename);
      saveParameters();
      setTitle(PRODUCT_NAME + " - " + filename);
      addRecentFileMenu();
    }

    // Save memory to hex file
    if (cmd.equals(BootloaderResourceBundle.getString("SAVE_HEX_FILE_TEXT"))) {
      if (parameters.getProgramName().equals("")) {
        saveAs();
      }
      else {
        f = new File(parameters.getProgramDir() + 
                     File.separator +
                     parameters.getProgramName());
        try {
          FileOutputStream fos = new FileOutputStream(f);
          saveHexFile(fos);
          fos.close();
        }
        catch (IOException e1) {
          System.err.println(e1);
        }
      }
    }

    // Open recent file
    switch(cmd.charAt(0)) {
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
      dir = parameters.getRecentProgramDir(cmd.charAt(0) - '0' - 1);
      filename = parameters.getRecentProgramName(cmd.charAt(0) - '0' - 1);
      f = new File(dir + File.separator + filename);
      
      try {
        FileInputStream fis = new FileInputStream(f);
        lastAddress = getLastAddress(fis);
        fis.close();
        resizeMemory(lastAddress);
        eraseMemory();
        fis = new FileInputStream(f);
        loadHexFile(fis);
        fis.close();
      }
      catch (FileNotFoundException e1) {
        AlertDialog ad = new AlertDialog(this, 
                                         BootloaderResourceBundle.getString("ERROR_OPENING_FILE_TEXT") + "!",
                                         BootloaderResourceBundle.getString("FILE_TEXT") + " " + filename + " " + BootloaderResourceBundle.getString("IN_DIRECTORY_TEXT"),
                                         dir,
				                         BootloaderResourceBundle.getString("NOT_FOUND_TEXT") + ".");

        return;
      }
      catch (IOException e2) {
        System.err.println(e2);
        return;
      }
      
      refreshMemory();
      refreshTimer.start();
      parameters.setProgramDir(dir);
      parameters.setProgramName(filename);
      parameters.setRecentProgram(dir, filename);
      saveParameters();
      setTitle(PRODUCT_NAME + " - " + filename);
      addRecentFileMenu();
      break;
    }

    // Save memory to hex file
    if (cmd.equals(BootloaderResourceBundle.getString("SAVE_HEX_FILE_AS_TEXT") + "...")) {
      saveAs();
    }
 
    // Calls shutdown, which exits the program.
    if (cmd.equals(BootloaderResourceBundle.getString("EXIT_TEXT"))) {
      shutdown();
    }

    // Edit Config
    if (cmd.equals(BootloaderResourceBundle.getString("CONFIG_TEXT"))) {
      ConfigDialog cd = new ConfigDialog(this, connection);
      refreshMemory();
      refreshTimer.start();
    }

    // Update EEPROM data in  memory
    if (cmd.equals(BootloaderResourceBundle.getString("EEPROM_DATA_TEXT"))) {
      EeDataDialog cd = new EeDataDialog(this, connection);
      refreshMemory();
      refreshTimer.start();
    }

    // Read PIC device
    if (cmd.equals(BootloaderResourceBundle.getString("READ_ALL_TEXT"))) {
      previousCursor = getCursor();
      setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
      configurationPanel.setParameters();
      
      setProgressStatus(BootloaderResourceBundle.getString("READING_DEVICE_TEXT"));

      readDeviceTransaction();
    }

    // Program PIC device
    if (cmd.equals(BootloaderResourceBundle.getString("PROGRAM_TEXT"))) {
      previousCursor = getCursor();
      setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
      if (parameters.getProgramAutoload())
        reloadHexFile();

      setProgressStatus(BootloaderResourceBundle.getString("PROGRAMMING_DEVICE_TEXT"));

      programDeviceTransaction();
    }

    // Write EEPROM Data
    if (cmd.equals(BootloaderResourceBundle.getString("PROGRAM_EEPROM_DATA_TEXT"))) {
      previousCursor = getCursor();
      setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

      setProgressStatus(BootloaderResourceBundle.getString("WRITING_DEVICE_EEPROM_DATA_TEXT"));

      writeEeDataTransaction();
    }

    // Write Config
    if (cmd.equals(BootloaderResourceBundle.getString("PROGRAM_CONFIG_TEXT"))) {
      previousCursor = getCursor();
      setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

      setProgressStatus(BootloaderResourceBundle.getString("WRITING_DEVICE_CONFIG_TEXT"));

      writeConfigTransaction();
    }

    // Verify that PIC device matches memory
    if (cmd.equals(BootloaderResourceBundle.getString("VERIFY_TEXT"))) {
      previousCursor = getCursor();
      setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

      setProgressStatus(BootloaderResourceBundle.getString("VERIFY_DEVICE_TEXT"));

      verifyDeviceTransaction();
    }

    // Send command to PIC device to get out of boot mode
    if (cmd.equals(BootloaderResourceBundle.getString("RUN_TEXT"))) {
      previousCursor = getCursor();
      setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

      setProgressStatus(BootloaderResourceBundle.getString("RUN_DEVICE_TEXT"));

      runDeviceTransaction();
    }

    // Set log level to debug
    if (cmd.equals("Debug")) {
      debugItem.setSelected(true);
      parameters.setLogLevel(Integer.toString(Log.DEBUG));
      log.setSeverityLevel(parameters.getLogLevel());
      saveParameters();
    }

    // Set log level to info
    if (cmd.equals("Info")) {
      infoItem.setSelected(true);
      parameters.setLogLevel(Integer.toString(Log.INFO));
      log.setSeverityLevel(parameters.getLogLevel());
      saveParameters();
    }

    // Set log level to error
    if (cmd.equals("Error")) {
      errorItem.setSelected(true);
      parameters.setLogLevel(Integer.toString(Log.ERROR));
      log.setSeverityLevel(parameters.getLogLevel());
      saveParameters();
    }

    // Set log level to none
    if (cmd.equals("None")) {
      noneItem.setSelected(true);
      parameters.setLogLevel(Integer.toString(Log.NONE));
      log.setSeverityLevel(parameters.getLogLevel());
      saveParameters();
    }

    if (cmd.equals(BootloaderResourceBundle.getString("DELAYED_TEXT"))) {
      parameters.setBootMode(BootloaderParameters.BOOT_MODE_DELAYED);
      
	  programResetItem.setEnabled(true);
	  
      saveParameters();
    }

    if (cmd.equals(BootloaderResourceBundle.getString("USER_DEFINED_TEXT"))) {
      parameters.setBootMode(BootloaderParameters.BOOT_MODE_USER);
      
      // Force reset after program item to false and disable menu item
	  parameters.setProgramReset("false");
	  programResetItem.setState(parameters.getProgramReset());
	  programResetItem.setEnabled(false);
	  
      saveParameters();
    }

    // Set autoload program indicator
    if (cmd.equals(BootloaderResourceBundle.getString("RELOAD_BEFORE_PROGRAM_TEXT"))) {
      boolean state = programAutoloadItem.getState();
      parameters.setProgramAutoload(Boolean.toString(state));
      saveParameters();
    }

    // Set reset program indicator
    if (cmd.equals(BootloaderResourceBundle.getString("RESET_AFTER_PROGRAM_TEXT"))) {
      boolean resetState = programResetItem.getState();
      parameters.setProgramReset(Boolean.toString(resetState));
      saveParameters();
    }

    // Set program Config indicator
    if (cmd.equals(BootloaderResourceBundle.getString("ENABLE_PROGRAM_CONFIG_TEXT"))) {
      boolean state = programConfigItem.getState();
      parameters.setProgramConfig(Boolean.toString(state));
      saveParameters();
    }

    // Set program EE Data indicator
    if (cmd.equals(BootloaderResourceBundle.getString("ENABLE_PROGRAM_EEPROM_DATA_TEXT"))) {
      boolean state = programEeDataItem.getState();
      parameters.setProgramEeData(Boolean.toString(state));
      saveParameters();
    }

	if (cmd.equals(BootloaderResourceBundle.getString("ENGLISH_TEXT"))) {
	  parameters.setLanguage(BootloaderParameters.LANGUAGE_ENGLISH);
	  saveParameters();
	  JOptionPane.showMessageDialog(null, BootloaderResourceBundle.getString("LANGUAGE_SELECTION_TEXT"),
									BootloaderResourceBundle.getString("CHANGE_LANGUAGE_TEXT"),
									JOptionPane.INFORMATION_MESSAGE);
	  System.exit(0);
	}

	if (cmd.equals(BootloaderResourceBundle.getString("FRENCH_TEXT"))) {
	  parameters.setLanguage(BootloaderParameters.LANGUAGE_FRENCH);
	  saveParameters();
	  JOptionPane.showMessageDialog(null, BootloaderResourceBundle.getString("LANGUAGE_SELECTION_TEXT"),
		                            BootloaderResourceBundle.getString("CHANGE_LANGUAGE_TEXT"),
									JOptionPane.INFORMATION_MESSAGE);
	  System.exit(0);
    }

    // Show the help dialog
    if (cmd.equals(BootloaderResourceBundle.getString("CONTENTS_TEXT"))) {
      HelpDialog hd = new HelpDialog(this);
    }

    // Show the about dialog
    if (cmd.equals(BootloaderResourceBundle.getString("ABOUT_TEXT") + " " + PRODUCT_NAME + "...")) {
      AboutDialog ad = new AboutDialog(this);
    }

    // Sends bytes to the port.
    if (cmd.equals("Send")) {
      boolean closed = false;

      if (!connection.isOpen()) {
        // First, open connection
        Cursor previousCursor = getCursor();

        closed = true;

        setNewCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
        configurationPanel.setParameters();

        try {
          connection.openConnection();
        }
        catch (PortInUseException e2) {
          AlertDialog ad = new AlertDialog(this, 
                                           BootloaderResourceBundle.getString("ERROR_OPENING_PORT_TEXT") + "!",
                                           BootloaderResourceBundle.getString("PORT_TEXT") + " " + parameters.getPortName() +
                                           " " + BootloaderResourceBundle.getString("IN_USE_TEXT") + ".",
                                           "",
                                           BootloaderResourceBundle.getString("RELEASE_PORT_TEXT") + ".");
          setNewCursor(previousCursor);
          return;
        }
        catch (SerialConnectionException e2) {
          AlertDialog ad = new AlertDialog(this, 
                                           BootloaderResourceBundle.getString("ERROR_OPENING_PORT_TEXT") + "!",
                                           "Error opening port,",
                                           e2.getMessage() + ".",
                                           "Select new settings, try again.");
          setNewCursor(previousCursor);
          return;
        }
        setNewCursor(previousCursor);
      }

      connection.sendBytes(sendLine.getText());

      if (closed) {
        // If connection was closed before we sent the bytes, then close it back
        connection.closeConnection();
      }
    }
  } // actionPerformed

  private void setProgressStatus(String title) {
  	if (progressDialog == null)
  	  return;
  	
  	progressDialog.setTitle(title);
    progressDialog.setVisible(true);
  } // setProgressStatus

  private void hideProgressDialog() {
  	if (progressDialog == null)
  	  return;
  	
    progressDialog.setVisible(false);
  } // hideProgressDialog
  
  public void reportProgress(int percent) {
    if (percent >= lastProgress + 10) {
      lastProgress = percent;
      if (progressDialog == null)
    	System.out.print(".");
    }
    	
  	if (progressDialog == null)
    	  return;

  	progressDialog.setProgressBar(percent);
  } // reportProgress
  
  private void cliReportChar(char c) {
  	System.out.print(c);
  } // cliReportStatus
  
  private void cliReportStatus(String status) {
  	System.out.println(status);
  } // cliReportStatus
  
  private boolean isCli() {
  	return cliFilename != null;
  } // isCli
  
  public void saveAs() {
    String dir;
    String filename;
    File f;

    if (parameters.getFileDialogUseSwing()) {
      JFileChooser chooser = new JFileChooser();
      BasicFileFilter filter = new BasicFileFilter();
      filter.addExtension("hex");
      filter.setDescription(BootloaderResourceBundle.getString("HEX_FILES_TEXT"));
      chooser.setFileFilter(filter);
      if (!parameters.getProgramDir().equals(""))
        chooser.setCurrentDirectory(new File(parameters.getProgramDir()));
      int returnVal = chooser.showSaveDialog(this);
      if (returnVal != JFileChooser.APPROVE_OPTION)
        return;
      filename = chooser.getSelectedFile().getName();
      dir = chooser.getCurrentDirectory().getPath();
      f = chooser.getSelectedFile();
    }
    else {
      FileDialog fd = new FileDialog(this, 
			                         BootloaderResourceBundle.getString("SAVE_HEX_FILE_TEXT"),
                                     FileDialog.SAVE);
      if (!parameters.getProgramDir().equals(""))
        fd.setDirectory(parameters.getProgramDir());
      fd.setFilenameFilter(new HexFilenameFilter());
      fd.setVisible(true);
      filename = fd.getFile();
      if (filename == null)
        return;
      dir = fd.getDirectory();
      f = new File(dir + filename);
    }

    try {
      FileOutputStream fos = new FileOutputStream(f);
      saveHexFile(fos);
      fos.close();
    } catch (IOException e1) {
      System.err.println(e1);
    }
    parameters.setProgramDir(dir);
    parameters.setProgramName(filename);
    saveParameters();
    setTitle(PRODUCT_NAME + " - " + parameters.getProgramName());
  } // saveAs

  public void commandTerminated(int reason) {
    AlertDialog ad;

    switch(reason) {
    case COMMAND_TERM_REASON_PORT_BUSY:
      if (cliFilename != null) {
      	System.err.println(BootloaderResourceBundle.getString("ERROR_OPENING_PORT_TEXT") + "! " +
      			           BootloaderResourceBundle.getString("PORT_TEXT") + " " + parameters.getPortName() + " " + BootloaderResourceBundle.getString("IN_USE_TEXT") + ".");
        break;
      }
      
      ad = new AlertDialog(this, 
                           BootloaderResourceBundle.getString("ERROR_OPENING_PORT_TEXT") + "!",
                           BootloaderResourceBundle.getString("PORT_TEXT") + " " + parameters.getPortName() + " " + BootloaderResourceBundle.getString("IN_USE_TEXT") + ".",
                           "",
			               BootloaderResourceBundle.getString("RELEASE_PORT_TEXT") + ".");
      break;
    case COMMAND_TERM_REASON_PORT_OPEN_FAILURE:
      if (cliFilename != null) {
        System.err.println(BootloaderResourceBundle.getString("ERROR_OPENING_PORT_TEXT") + "! " +
                           BootloaderResourceBundle.getString("ERROR_OPENING_PORT_LOWER_TEXT"));
        break;
      }
      
      ad = new AlertDialog(this, 
                           BootloaderResourceBundle.getString("ERROR_OPENING_PORT_TEXT") + "!",
                           BootloaderResourceBundle.getString("ERROR_OPENING_PORT_LOWER_TEXT") + ",",
                           "",
                           BootloaderResourceBundle.getString("SELECT_NEW_SETTINGS_TEXT") + ".");
      break;
    case COMMAND_TERM_REASON_TIMEOUT:
      if (cliFilename != null) {
        System.err.println(BootloaderResourceBundle.getString("COMMUNICATIONS_ERROR_TEXT") + "! " +
                           BootloaderResourceBundle.getString("TIMEOUT_TEXT"));
        break;
      }
      
      ad = new AlertDialog(this, 
                           BootloaderResourceBundle.getString("COMMUNICATIONS_ERROR_TEXT") + "!",
                           BootloaderResourceBundle.getString("COMMUNICATIONS_ERROR_TEXT") + ":",
                           BootloaderResourceBundle.getString("TIMEOUT_TEXT") + ".",
                           BootloaderResourceBundle.getString("TRY_AGAIN_TEXT") + ".");
      break;
	case COMMAND_TERM_REASON_RESOURCE_NOT_AVAILABLE:
    case COMMAND_TERM_REASON_USER:
    case COMMAND_TERM_REASON_VERIFY_FAILURE:
	case COMMAND_TERM_REASON_UNEXPECTED_MESSAGE:
      break;
    }
    
    hideProgressDialog();
    
    setNewCursor(previousCursor);
  } // commandTerminated

  public void reloadHexFile() {
    File f = new File(parameters.getProgramDir() + File.separator +
                      parameters.getProgramName());

    try {
      FileInputStream fis = new FileInputStream(f);
      eraseMemory();
      loadHexFile(fis);
      refreshMemory();
      refreshTimer.start();
      fis.close();
    } catch (FileNotFoundException e1) {
      // Skip reload. Program what is in memory.
      System.err.println("File " + 
                         parameters.getProgramDir() + File.separator +
                         parameters.getProgramName() + ": " + e1); 
      return;
    } catch (IOException e2) {
      System.err.println(e2);
    }
  } // reloadHexFile

  /**
  Sets the <code>Cursor</code> for the application.
  @param c New <code>Cursor</code>
  */
  private void setNewCursor(Cursor c) {
    setCursor(c);
    messageAreaIn.setCursor(c);
    memoryArea.setCursor(c);
  } // setNewCursor

  /**
  Cleanly shuts down the application. first closes any open ports and
  cleans up, then exits.
  */
  private void shutdown() {
    connection.closeConnection();
    System.exit(EXIT_NORMAL);
  } // shutdown

  /**
  Load PIC device table.
  */
  private void loadDeviceTable() {
    deviceTable = new PicDeviceTable();

    deviceTable.load();
  } // loadDeviceProperties

  /**
  Get PIC device associated with device with specified Device Id.

  @param deviceId Device ID
  */
  public PicDevice getPicDevice(int deviceId) {
    return (PicDevice) deviceTable.get(deviceId);
  } // getPicDevice

  /**
  Writes the current parameters to a configuration file of the 
  java.properties style.

  @param path Directory and filename of Bootloader property file.
  */
  private void writeFile(String path) {
    Properties newProps;
    FileOutputStream fileOut = null;

    newProps = new Properties();

	newProps.put("language", parameters.getLanguageString());
	newProps.put("portName", parameters.getPortName());
    newProps.put("baudRate", parameters.getBaudRateString());
    newProps.put("timeoutDelay", parameters.getTimeoutDelayString());
    newProps.put("maxTimeouts", parameters.getMaxTimeoutsString());
    newProps.put("maxPacketSize", parameters.getMaxPacketSizeString());
    newProps.put("logLevel", parameters.getLogLevelString());
    newProps.put("bootMode", parameters.getBootModeString());
    newProps.put("programAutoload", parameters.getProgramAutoloadString());
    newProps.put("programReset", parameters.getProgramResetString());
    newProps.put("programConfig", parameters.getProgramConfigString());
    newProps.put("programEeData", parameters.getProgramEeDataString());
    newProps.put("programDir", parameters.getProgramDir());
    newProps.put("programName", parameters.getProgramName());
	newProps.put("fileDialogUseSwing", parameters.getFileDialogUseSwingString());
	newProps.put("pic18fxx2Ind", parameters.getPic18fxx2IndString());
    if (parameters.getRecentProgramDir(0) != null)
      newProps.put("recentProgramDir0", parameters.getRecentProgramDir(0));
    if (parameters.getRecentProgramName(0) != null)
      newProps.put("recentProgramName0", parameters.getRecentProgramName(0));
    if (parameters.getRecentProgramDir(1) != null)
      newProps.put("recentProgramDir1", parameters.getRecentProgramDir(1));
    if (parameters.getRecentProgramName(1) != null)
      newProps.put("recentProgramName1", parameters.getRecentProgramName(1));
    if (parameters.getRecentProgramDir(2) != null)
      newProps.put("recentProgramDir2", parameters.getRecentProgramDir(2));
    if (parameters.getRecentProgramName(2) != null)
      newProps.put("recentProgramName2", parameters.getRecentProgramName(2));
    if (parameters.getRecentProgramDir(3) != null)
      newProps.put("recentProgramDir3", parameters.getRecentProgramDir(3));
    if (parameters.getRecentProgramName(3) != null)
      newProps.put("recentProgramName3", parameters.getRecentProgramName(3));
    if (parameters.getRecentProgramDir(4) != null)
      newProps.put("recentProgramDir4", parameters.getRecentProgramDir(4));
    if (parameters.getRecentProgramName(4) != null)
      newProps.put("recentProgramName4", parameters.getRecentProgramName(4));

    try {
      fileOut = new FileOutputStream(path);
    } catch (IOException e) {
      System.out.println("Could not open file for writing");
    }

    try {
      newProps.store(fileOut, "Bootloader properties");
    } catch (IOException e) {
      System.out.println("Could not store properties: " + e);
    }

    try {
      fileOut.close();
    } catch (IOException e1) {
      System.out.println("Could not close file for writing");
    }
  } // writeFile

  /**
  Finds configuration file in arguments and creates a properties object from
  that file.
  */
  private void loadParameters() {
    File f = new File("bootloader.properties");

    if (!f.exists()) {
      f = new File(System.getProperty("user.dir") + File.separator +
                   "bootloader.properties");
    }

    if (f.exists()) {
      try {
        FileInputStream fis = new FileInputStream(f);
        props = new Properties();
        props.load(fis);
        fis.close();
        loadParams();
      } catch (IOException e) {
        System.err.println(e);
      }
    }
    else {
      saveParameters();
    }
  } // loadParameters

  /**
  Set the parameters object to the settings in the properties object.
  */
  private void loadParams() {
	parameters.setLanguage(props.getProperty("language"));
    parameters.setPortName(props.getProperty("portName"));
    parameters.setBaudRate(props.getProperty("baudRate"));
    parameters.setTimeoutDelay(props.getProperty("timeoutDelay"));
    parameters.setMaxTimeouts(props.getProperty("maxTimeouts"));
    parameters.setMaxPacketSize(props.getProperty("maxPacketSize"));
    parameters.setLogLevel(props.getProperty("logLevel"));
    parameters.setBootMode(props.getProperty("bootMode"));
    parameters.setProgramAutoload(props.getProperty("programAutoload"));
    parameters.setProgramReset(props.getProperty("programReset"));
    parameters.setProgramConfig(props.getProperty("programConfig"));
    parameters.setProgramEeData(props.getProperty("programEeData"));
    parameters.setProgramDir(props.getProperty("programDir"));
    parameters.setProgramName(props.getProperty("programName"));
	parameters.setFileDialogUseSwing(props.getProperty("fileDialogUseSwing"));
	parameters.setPic18fxx2Ind(props.getProperty("pic18fxx2Ind"));
    parameters.setFlowControlIn("None");
    parameters.setFlowControlOut("None");
    parameters.setParity("None");
    parameters.setDatabits("8");
    parameters.setStopbits("1");
    if (props.getProperty("recentProgramName4") != null) {
      parameters.setRecentProgram(props.getProperty("recentProgramDir4"),
                                  props.getProperty("recentProgramName4"));
    }
    if (props.getProperty("recentProgramName3") != null) {
      parameters.setRecentProgram(props.getProperty("recentProgramDir3"),
                                  props.getProperty("recentProgramName3"));
    }
    if (props.getProperty("recentProgramName2") != null) {
      parameters.setRecentProgram(props.getProperty("recentProgramDir2"),
                                  props.getProperty("recentProgramName2"));
    }
    if (props.getProperty("recentProgramName1") != null) {
      parameters.setRecentProgram(props.getProperty("recentProgramDir1"),
                                  props.getProperty("recentProgramName1"));
    }
    if (props.getProperty("recentProgramName0") != null) {
      parameters.setRecentProgram(props.getProperty("recentProgramDir0"),
                                  props.getProperty("recentProgramName0"));
    }
  } // loadParams

  public void saveParameters() {
    writeFile("bootloader.properties");
  } // saveParameters

  public String getProgramDir() {
    return props.getProperty("programDir");
  } // getProgramDir

  /**
  GUI element that holds the user changeable elements for connection
  configuration.
  */
  class ConfigurationPanel extends JPanel implements ItemListener {
    private Bootloader parent;

    private JLabel portNameLabel;
    private Choice portChoice;

    private JLabel baudLabel;
    private Choice baudChoice;

    /**
    Creates and initilizes the configuration panel. The initial settings
    are from the parameters object.
    */
    public ConfigurationPanel(Bootloader parent) {
      this.parent = parent;

      setLayout(new GridLayout(1, 2));

      portNameLabel = new JLabel(BootloaderResourceBundle.getString("PORT_NAME_TEXT") + ":", Label.LEFT);
      add(portNameLabel);

      portChoice = new Choice();
      portChoice.addItemListener(this);
      add(portChoice);
      listPortChoices();
      portChoice.select(parameters.getPortName());
	    
      baudLabel = new JLabel("Baud Rate:", Label.LEFT);
      add(baudLabel);

      baudChoice = new Choice();
      baudChoice.add("9600");
      baudChoice.add("19200");
      baudChoice.add("38400");
      baudChoice.add("57600");
      baudChoice.add("115200");
      baudChoice.select(Integer.toString(parameters.getBaudRate()));
      baudChoice.addItemListener(this);
      add(baudChoice);
    } // ConfigurationPanel

    /**
    Sets the configuration panel to the settings in the parameters object.
    */
    public void setConfigurationPanel() {
      portChoice.select(parameters.getPortName());
      baudChoice.select(parameters.getBaudRateString());
    } // setConfigurationPanel

    /**
    Sets the parameters object to the settings in the configuration panel.
    */
    public void setParameters() {
      parameters.setPortName(portChoice.getSelectedItem());
      parameters.setBaudRate(baudChoice.getSelectedItem());
      parameters.setFlowControlIn("None");
      parameters.setFlowControlOut("None");
      parameters.setDatabits("8");
      parameters.setStopbits("1");
      parameters.setParity("None");
    } // setParameters

    /**
    Sets the elements for the portChoice from the ports available on the
    system. Uses an emuneration of comm ports returned by 
    CommPortIdentifier.getPortIdentifiers(), then sets the current
    choice to a mathing element in the parameters object.
    */
    void listPortChoices() {
      CommPortIdentifier portId;

      Enumeration en = CommPortIdentifier.getPortIdentifiers();

      // Iterate through the ports.
      while (en.hasMoreElements()) {
        portId = (CommPortIdentifier) en.nextElement();
        if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
          portChoice.addItem(portId.getName());
        }
      }
      portChoice.select(parameters.getPortName());
    } // listPortChoices

    /**
    Event handler for changes in the current selection of the Choices.
    If a port is open the port can not be changed.
    If the choice is unsupported on the platform then the user will
    be notified and the settings will revert to their pre-selection
    state.
    */
    public void itemStateChanged(ItemEvent e) {
      // Check if port is open.
      if (connection.isOpen()) {
        // If port is open do not allow port to change.
        if (e.getItemSelectable() == portChoice) {
          // Alert user.
          AlertDialog ad = new AlertDialog(parent, "Port Open!",
                                           "Port can not", 
                                           "be changed",
                                           "while a port is open.");

          // Return configurationPanel to pre-choice settings.
          setConfigurationPanel();
          return;
        }
        // Set the parameters from the choice panel.
        setParameters();
        parent.saveParameters();
        try {
          // Attempt to change the settings on an open port.
          connection.setConnectionParameters();
        } catch (SerialConnectionException ex) {
          // If setting can not be changed, alert user, return to
          // pre-choice settings.
          AlertDialog ad = new AlertDialog(parent, 
                                           "Unsupported Configuration!",
                                           "Configuration Parameter unsupported,",
                                           "select new value.",
                                           "Returning to previous configuration.");
          setConfigurationPanel();
        }
      } else {
        // Since port is not open just set the parameter object.
        setParameters();
        parent.saveParameters();
      }
    } // itemStateChanged
  } // ConfigurationPanel

  /**
  Handles closing down system. Allows application to be closed with window
  close box.
  */
  class CloseHandler extends WindowAdapter {
    Bootloader sd;
	
    public CloseHandler(Bootloader sd) {
      this.sd = sd;
    }

    public void windowClosing(WindowEvent e) {
      sd.shutdown();
    }
  } // CloseHandler

  private int getLastAddress(FileInputStream f) {
    int offset;
    int dataCount;
    int code;
    int address;
    byte line[];
    int lineAddrHigh;
    int lastAddress = 0;

    line = new byte[32];

    lineAddrHigh = 0;

    for ( ; ; ) {
      if (!readNextLine(f, line))
        break;

      offset = 0;

      // Read :
      if (line[offset++] != ':') {
        continue;
      }

      // Read data count (1 byte)
      dataCount = line[offset++];
      // Read address (2 bytes)
      address = ((((int) line[offset]) & 0xff) << 8) | 
        (((int) line[offset + 1]) & 0xff);
      offset += 2;
      // Read code (1 byte)
      code = line[offset++];
      switch(code) {
      case CODE_DATA:
        if (lineAddrHigh < 32) {
          // Read data (1 byte times data count)
          address +=  lineAddrHigh * 0x10000;
          if (address + dataCount > lastAddress)
            lastAddress = address + dataCount - 1;
        }
        break;
      case CODE_EOF:
        return lastAddress;
      case CODE_ADDR_HI:
        lineAddrHigh = ((((int) line[offset]) & 0xff) << 8) |
          (((int) line[offset + 1]) & 0xff);
        break;
      }
    }

    return lastAddress;
  } // getLastAddress

  private void loadHexFile(FileInputStream f) {
    int offset;
    int dataCount;
    int code;
    int address;
    byte line[];
    int lineAddrHigh;
    int i;

    line = new byte[32];

    lineAddrHigh = 0;

    for ( ; ; ) {
      if (!readNextLine(f, line))
        break;

      offset = 0;

      // Read :
      if (line[offset++] != ':') {
        continue;
      }

      // Read data count (1 byte)
      dataCount = line[offset++];
      // Read address (2 bytes)
      address = ((((int) line[offset]) & 0xff) << 8) |
        (((int) line[offset + 1]) & 0xff);
      offset += 2;
      // Read code (1 byte)
      code = line[offset++];
      switch(code) {
      case CODE_DATA:
        if (lineAddrHigh < 32) {
          // Read data (1 byte times data count)
          address +=  lineAddrHigh * 0x10000;
          if (address < 0x0200) {
            // Do not process data that overwrites bootloader code space
            log.log(Log.DEBUG, "Skip instructions at address " + address + "\n");
            continue;
          }
          for (i = 0; i < dataCount; i++) {
            loadProgramMem(address + i, line[offset + i]);
          }
        }
        else if (lineAddrHigh == 32) { // User memory
        }
        else if (lineAddrHigh == ADDR_HI_CONFIG) { // Config
          address = lineAddrHigh * 0x10000 + (address & 0xffff); // was (address & 0xfff0)
          for (i = 0; i < dataCount; i++) {
            loadConfig(address + i, line[offset + i]);
          }
        }
        else if (lineAddrHigh == ADDR_HI_EE_DATA) { // EE Data
          for (i = 0; i < dataCount; i++) {
            loadData(address + i, line[offset + i]);
          }
        }
        break;
      case CODE_EOF:
        return;
      case CODE_ADDR_HI:
        lineAddrHigh = ((((int) line[offset]) & 0xff) << 8) |
          (((int) line[offset + 1]) & 0xff);
        break;
      }
    }
  } // loadHexFile

  public boolean readNextLine(FileInputStream f, byte line[]) {
    int len = 0;
    StringBuffer data;
    int header, hi, lo;

    try {
      header = f.read();
    } catch (IOException e2) {
      System.err.println(e2);
      return false;
    }

    if (header == -1)
      return false;

    line[len++] = (byte) header;

    if (header != ':') {
      return true;
    }

    data = new StringBuffer("00");

    for ( ; ; ) {
      try {
        hi = f.read();
      } catch (IOException e2) {
        System.err.println(e2);
        return false;
      }

      // Check for end of file
      if (hi == -1) {
        if (len != 0)
          break;

        return false;
      }

      // Check for end of line
      if (!Utils.isHexDigit(hi))
        break;

      try {
        lo = f.read();
      } catch(IOException e2) {
        System.err.println(e2);
        return false;
      }

      if (!Utils.isHexDigit(lo)) {
        data.setCharAt(0, (char) hi);
        data.delete(1, 1);
        line[len++] = (byte) Utils.asciiToHex(data);
        return true;
      }

      data.setCharAt(0, (char) hi);
      data.setCharAt(1, (char) lo);
      line[len++] = (byte) Utils.asciiToHex(data);
    }

    return true;
  } // readNextLine

  private void saveHexFile(FileOutputStream f) {
    PicDevice device;
    int addr;
    int i;
    int checksum;
    int dataCount;
    int data;
    int lastAddress;

    device = getDevice();
    lastAddress = getProgramLastAddress();

    try {
      // Store program memory
      for (addr = device.loProgramMem; addr <= lastAddress; addr += DATA_COUNT) {
        checksum = 0;
        f.write(new String(":").getBytes());
        // Write data count
        if (lastAddress - addr + 1 > DATA_COUNT)
          dataCount = DATA_COUNT;
        else
          dataCount = lastAddress - addr + 1;
        f.write(Utils.hexToAscii(dataCount).getBytes());
        checksum += dataCount;

        // Write address high
        f.write(Utils.hexToAscii((addr >> 8) & 0xff).getBytes());
        checksum += (addr >> 8) & 0xff;
        // Write address low
        f.write(Utils.hexToAscii(addr & 0xff).getBytes());
        checksum += addr & 0xff;
        // Write code
        f.write(Utils.hexToAscii(CODE_DATA).getBytes());
        checksum += CODE_DATA;
        // Write data
        for (i = 0; i < dataCount; i++) {
          data = getProgramMem(addr + i);
          f.write(Utils.hexToAscii(data).getBytes());
          checksum += data;
        }
        // Write checksum
        f.write(Utils.hexToAscii((~(checksum) + 1) & 0xff).getBytes()); 
        f.write(new String("\r\n").getBytes());
      }

      // Store config
      if (isConfigPresent()) {
        checksum = 0;

        // First, write the new base address
        f.write(new String(":").getBytes());
        // Write data count
        f.write(Utils.hexToAscii(2).getBytes());
        checksum += 2;
        // Write address
        f.write(Utils.hexToAscii(0).getBytes());
        f.write(Utils.hexToAscii(0).getBytes());
         // Write code
        f.write(Utils.hexToAscii(CODE_ADDR_HI).getBytes());
        checksum += CODE_ADDR_HI;   
        // Write address high
        f.write(Utils.hexToAscii((device.loConfigMem >> 24) & 0xff).getBytes());
        checksum += (device.loConfigMem >> 24) & 0xff;
        f.write(Utils.hexToAscii((device.loConfigMem >> 16) & 0xff).getBytes());
        checksum += (device.loConfigMem >> 16) & 0xff;
        // Write checksum
        f.write(Utils.hexToAscii((~(checksum) + 1) & 0xff).getBytes());
        f.write(new String("\r\n").getBytes());

        // Then, write the config data
        f.write(new String(":").getBytes());
        checksum = 0;
        // Write data count
        dataCount = device.hiConfigMem - device.loConfigMem + 1;
        f.write(Utils.hexToAscii(dataCount).getBytes());
        checksum += dataCount;
        // Write address
        addr = device.loConfigMem & 0xffff;
        // Write address high
        f.write(Utils.hexToAscii((addr >> 8) & 0xff).getBytes());
        checksum += (addr >> 8) & 0xff;
        // Write address low
        f.write(Utils.hexToAscii(addr & 0xff).getBytes());
        checksum += addr & 0xff;
        // Write code
        f.write(Utils.hexToAscii(CODE_DATA).getBytes());
        checksum += CODE_DATA;
        // Write data
        for (addr = device.loConfigMem ; addr <= device.hiConfigMem; addr++) {
          data = getConfig(addr);
          f.write(Utils.hexToAscii(data).getBytes());
          checksum += data;
        }
        // Write checksum
        f.write(Utils.hexToAscii((~(checksum) + 1) & 0xff).getBytes());
        f.write(new String("\r\n").getBytes());
      }

      // Store EEPROM Data
      if (isEeDataPresent()) {
        checksum = 0;

        // First, write the new base address
        f.write(new String(":").getBytes());
        // Write data count
        f.write(Utils.hexToAscii(2).getBytes());
        checksum += 2;
        // Write address
        f.write(Utils.hexToAscii(0).getBytes());
        f.write(Utils.hexToAscii(0).getBytes());
         // Write code
        f.write(Utils.hexToAscii(CODE_ADDR_HI).getBytes());
        checksum += CODE_ADDR_HI;   
        // Write address high
        f.write(Utils.hexToAscii((ADDR_HI_EE_DATA >> 8) & 0xff).getBytes());
        checksum += (ADDR_HI_EE_DATA >> 8) & 0xff;
        f.write(Utils.hexToAscii(ADDR_HI_EE_DATA).getBytes());
        checksum += ADDR_HI_EE_DATA;
        // Write checksum
        f.write(Utils.hexToAscii((~(checksum) + 1) & 0xff).getBytes());
        f.write(new String("\r\n").getBytes());

        // Then, write the EEPROM Data
        dataCount = 16;
        for (addr = device.loEeData; addr <= device.hiEeData; addr += dataCount) {
          f.write(new String(":").getBytes());
          checksum = 0;
          // Write data count
          if (device.hiEeData - addr + 1 < 16)
            dataCount = device.hiEeData - addr + 1;
          f.write(Utils.hexToAscii(dataCount).getBytes());
          checksum += dataCount;
          // Write address
          // Write address high
          f.write(Utils.hexToAscii((addr >> 8) & 0xff).getBytes());
          checksum += (addr >> 8) & 0xff;
          // Write address low
          f.write(Utils.hexToAscii(addr & 0xff).getBytes());
          checksum += addr & 0xff;
          // Write code
          f.write(Utils.hexToAscii(CODE_DATA).getBytes());
          checksum += CODE_DATA;
          // Write data
          for (i = 0; i < dataCount; i++) {
            data = getEeData(addr + i);
            f.write(Utils.hexToAscii(data).getBytes());
            checksum += data;
          }
          // Write checksum
          f.write(Utils.hexToAscii((~(checksum) + 1) & 0xff).getBytes());
          f.write(new String("\r\n").getBytes());
        }
      }

      // Write EOF marker
      f.write(new String(":").getBytes());
      checksum = 0;
      // Write data count
      f.write(Utils.hexToAscii(0).getBytes());
      // Write address
      // Write address high
      f.write(Utils.hexToAscii(0).getBytes());
      // Write address low
      f.write(Utils.hexToAscii(0).getBytes());
      // Write code
      f.write(Utils.hexToAscii(CODE_EOF).getBytes());
      checksum += CODE_EOF;
      // Write checksum
      f.write(Utils.hexToAscii((~(checksum) + 1) & 0xff).getBytes());
      f.write(new String("\r\n").getBytes());
    } catch (IOException e) {
    }
  } // saveHexFile

  class HexFilenameFilter implements FilenameFilter {
    public boolean accept(File dir, String name) {
      return name.endsWith(".hex");
    }
  } // HexFilenameFilter
  
  public void processPacket(byte[] receiveBuffer) {
	log.log(Log.INFO, "Processing packet. Command: " +
			Utils.hexToAscii(receiveBuffer[COMMAND]) + "\n");

    retryTimer.stop();
    
	if (currentCommand == CommandNull) {
	  log.log(Log.ERROR, "Unexpected packet: dropped\n");
	  return;
	}
	
	switch(receiveBuffer[COMMAND]) {
	case CommandReadVersion:
	  recordFirmwareVersion(receiveBuffer);
	  readDeviceId();
	  break;
	case CommandReadProgramMemory:
	  if (currentOffset == 0x3ffffe) {
		recordDeviceId(receiveBuffer);
		resumeTransaction();
	  }
	  else {
		if (!verifyInd)
		  recordProgramMemory(receiveBuffer);
		else {
		  if (!verifyBlock(receiveBuffer)) {
			terminateCommand(COMMAND_TERM_REASON_VERIFY_FAILURE);
			break;
		  }
		}

	    readDeviceNext();
	  }
	  break;
	case CommandEraseProgramMemory:
	  eraseDeviceNext();
	  break;
	case CommandWriteProgramMemory:
	  programDeviceNext();
	  break;
	case CommandReadConfig:
	  recordConfig(receiveBuffer);
	  readConfigNext();
	  break;
	case CommandReadEeData:
	  recordEeData(receiveBuffer);
	  readEeDataNext();
	  break;
	case CommandWriteConfig:
	  writeConfigNext();
	  break;
	case CommandWriteEeData:
	  if (startupDelay)
	    resumeTransaction();
      else
        writeEeDataNext();
	  break;
	default:
	  log.log(Log.ERROR, "Command " +
			  Utils.hexToAscii(receiveBuffer[COMMAND]) + " not supported\n");
	}
  } // processPacket

  public void fillAsciiBuffer(int data) {
	if (data == SerialConnection.STX || data == SerialConnection.ETX || data == SerialConnection.DLE) {
	  lastCommandSent[lastCommandSentLength++] = SerialConnection.DLE;
	}

	lastCommandSent[lastCommandSentLength++] = (byte) data;
  } // fillAsciiBuffer

  public void sendCommand(int command, int dataLen, int address,
						  byte data[], int timeouts, int delay) {
	int checksum = 0;
	int addrl, addrh, addru;
	int i;
	int dLen = dataLen;

	timeout = 0;
	maxTimeouts = timeouts;
	if (delay != 0)
	  retryTimer.setDelay(delay);

	if (!connection.isOpen()) {
	  return;
	}

	log.log(Log.INFO, "Sending command " + command + "\n");

	lastCommandSentLength = 0;

	connection.sendRawByte((int) SerialConnection.STX); // STX
	lastCommandSent[lastCommandSentLength++] = SerialConnection.STX;
	connection.sendRawByte((int) SerialConnection.STX); // STX
	lastCommandSent[lastCommandSentLength++] = SerialConnection.STX;

	connection.sendData((int) command); // COMMAND
	checksum += command;
	fillAsciiBuffer(command);

	if (command == CommandWriteProgramMemory) {
	  dLen = dataLen / device.writeBlock; // Indicates the number of blocks
	  if (dataLen % device.writeBlock != 0)
	    log.log(Log.INFO, "WriteProgramMemory: data length in mot multiple of " + 
	            device.writeBlock);
	}
	
	connection.sendData((int) dLen); // DLEN
	checksum += dLen;
	fillAsciiBuffer(dLen);

	addrl = address & 0xff;
	connection.sendData((int) addrl); // ADDRL
	checksum += addrl;
	fillAsciiBuffer(addrl);

	addrh = (address & 0xff00) >> 8;
	connection.sendData((int) addrh); // ADDRH
	checksum += addrh;
	fillAsciiBuffer(addrh);

	addru = (address & 0xff0000) >> 16;
	connection.sendData((int) addru); // ADDRU
	checksum += addru;
	fillAsciiBuffer(addru);

	if (data != null) {
	  for (i = 0; i < dataLen; i++) {
	    connection.sendData((int) data[i]);
	    checksum += data[i];
	    fillAsciiBuffer(data[i]);
  	  }
	}

    checksum = (~checksum + 1) & 0xff;

	// Insert DLE before checksum
	connection.sendData((int) checksum); // CHECKSUM
	fillAsciiBuffer(checksum);

	connection.sendRawByte((int) SerialConnection.ETX); // ETX
	lastCommandSent[lastCommandSentLength++] = SerialConnection.ETX;

	if (log.getSeverityLevel() != Log.NONE) {
	  for (i = 0; i < lastCommandSentLength; i++)
		log.log(Log.DEBUG, Utils.hexToAscii(lastCommandSent[i]));
		log.log(Log.DEBUG, "\n");
    }

	if (delay != 0)
	  retryTimer.start();
  } // sendCommand

  public void retryCommand() {
	int i;
	String operationStr;
	int command;
	int offset = currentOffset;
	
	if (isCli())
		cliReportChar('T');
	
	timeout++;

	if (timeout == maxTimeouts) {
	  hideProgressDialog();
	  command = currentCommand;
	  operationStr = getOperationString();
	  terminateCommand(COMMAND_TERM_REASON_TIMEOUT);
	  AlertDialog ad = new AlertDialog(this, operationStr,
									   BootloaderResourceBundle.getString("TIMEOUT_WAITING_TEXT"),
									   BootloaderResourceBundle.getString("OPERATION_ABORTED_TEXT"),
									   "");
	  if (command == CommandReadVersion ||
		  (command == CommandReadProgramMemory && offset == 0x3ffffe)) {
		connection.scheduleCloseConnection();
	  }
	  return;
	}

	log.log(Log.INFO, "Retry command " + lastCommandSent[COMMAND + 2] +
			": timeout = " + timeout + "\n");

	for (i = 0; i < lastCommandSentLength; i++)
	  log.log(Log.DEBUG, Utils.hexToAscii(lastCommandSent[i]));
	log.log(Log.DEBUG, "\n");

	for (i = 0; i < lastCommandSentLength; i++) {
	  connection.sendRawByte((int) lastCommandSent[i]);
	}
	retryTimer.start();
  } // retryCommand

  private void connect() {
	try {
	  connection.openConnection();
	}
	catch (PortInUseException e1) {
	  commandTerminated(COMMAND_TERM_REASON_PORT_BUSY);
	  hideProgressDialog();
	  System.err.println("Error Opening Port! " + e1.getMessage());
	  return;
	}
	catch (SerialConnectionException e2) {
	  commandTerminated(COMMAND_TERM_REASON_PORT_OPEN_FAILURE);
	  hideProgressDialog();
	  System.err.println("Error Opening Port! " + e2.getMessage());
	  return;
	}

	if (isCli() && cliVerbose)
		cliReportStatus("Opened serial port");
	
	readVersion();
  } // connect

  private void resumeTransaction() {
	switch(currentTransaction) {
	case CommandReadProgramMemory:
	  if (!verifyInd)
		readDevice();
	  else
		verifyDevice();
	  break;
	case CommandWriteProgramMemory:
	  eraseDevice();
	  break;
	case CommandWriteEeData:
	  writeEeData(eeDataInd);
	  break;
	case CommandWriteConfig:
	  writeConfig();
	  break;
	case CommandReset:
	  if (startupDelay) {
	  	// We have written the startup delay on the last byte of EEPROM.
	  	// We can now send the reset command.
		resetDevice();
		terminateCommand(COMMAND_TERM_REASON_SUCCESS);
	  	startupDelay = false;
	  }
	  else
	    runDevice();
	  break;
	}
  } // resumeTransaction

  public void recordFirmwareVersion(byte[] receiveBuffer) {
	StringBuffer buffer = new StringBuffer();

	if (isCli() && cliVerbose)
	  cliReportStatus("\nEstablished communication with device");
	
	log.log(Log.DEBUG, "Record firmware version " +
			(int) receiveBuffer[VERSIONH] + "." +
			(int) receiveBuffer[VERSIONL] + "\n");

	firmwareVersion = (receiveBuffer[VERSIONH] << 8) |
	  receiveBuffer[VERSIONL];

	buffer.append((int) receiveBuffer[VERSIONH]);
	buffer.append('.');
	buffer.append((int) receiveBuffer[VERSIONL]);

	firmwareVersionField.setText(buffer.toString());

	currentCommand = CommandNull;
  } // recordFirmwareVersion

  public void readVersion() {
	int baudRate;
	byte data[] = new byte[1];

	if (isCli() && cliVerbose)
      cliReportStatus("Communicating with microcontroller");
	
	currentCommand = CommandReadVersion;

	currentOffset = device.hiEeData;

	if (parameters.getBootMode() == BootloaderParameters.BOOT_MODE_USER)
	  baudRate = 0xff;
	else
	  baudRate = (parameters.getBaudRate() + 500) / 1000;

	data[0] = (byte) baudRate;

	sendCommand(CommandReadVersion, 1, currentOffset, data, -1, 100);
  } // readVersion

  public void recordDeviceId(byte[] receiveBuffer) {    
    if (receiveBuffer[COMMAND] != currentCommand) {
      log.log(Log.ERROR, "Unexpected command " + receiveBuffer[COMMAND] + " received: " +
              "was expecting command " + currentCommand);
      terminateCommand(COMMAND_TERM_REASON_UNEXPECTED_MESSAGE);
      return;
    }
    
	deviceId = ((int) (receiveBuffer[DATA] & 0xff)) >> 5;
	deviceId += ((int) (receiveBuffer[DATA + 1] & 0xff)) << 3;

	deviceIdRaw[0] = (int) (receiveBuffer[DATA] & 0xff);
	deviceIdRaw[1] = (int) (receiveBuffer[DATA + 1] & 0xff);

	device = getPicDevice(deviceId);

	if (device == null) {
	  device = new PicDevice(BootloaderResourceBundle.getString("UNKNOWN_TEXT") + " ID#" + deviceId, 0x200, 0x7FFF, 0, 255,
					         "FOSC20,BORV,BOREN,PWRTEN,WDTPS,WDTEN,DEBUG,LVP,STVREN,CP1,CP0,CPD,CPB,WRT1,WRT0,WRTD,WRTB,WRTC,EBTR1,EBTR0,EBTRB");
	}

	deviceIdChoice.removeAll();
	
	if (device.alternateDevice != null) {
	  // There are more than one device associated with this Device ID.
	  PicDevice deviceXx39;
	  if (device.name.endsWith("39") || device.name.endsWith("90"))
	    deviceXx39 = device;
	  else {
	    deviceXx39 = device.alternateDevice;
	    deviceXx39.alternateDevice = device;
	  }
      deviceIdChoice.add(deviceXx39.alternateDevice.name);
	  deviceIdChoice.add(deviceXx39.name);
	  
	  if (parameters.getPic18fxx2Ind())
	  	device = deviceXx39.alternateDevice;
	  else {
	    deviceIdChoice.select(deviceXx39.name);
	    device = deviceXx39;
	  }
	}
    else
	  deviceIdChoice.add(device.name);

    // Make sure lastWriteOffset matches current connected device
    if (currentTransaction == CommandReadProgramMemory)
      lastWriteOffset = device.hiProgramMem;
    
	currentCommand = CommandNull;
  } // recordDeviceId
  
  public void readDeviceId() {
  	int length = 2;
  	
  	// Read Device ID field in PIC
	currentCommand = CommandReadProgramMemory;

	currentOffset = 0x3ffffe;

	sendCommand(CommandReadProgramMemory, length, currentOffset, null,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // readDeviceId
  	
  public void recordProgramMemory(byte[] receiveBuffer) {
	int baseAddress = ((((int) receiveBuffer[ADDRU]) & 0xff) << 16) +
	  ((((int) receiveBuffer[ADDRH]) & 0xff) << 8) +
	  (((int) receiveBuffer[ADDRL]) & 0xff);
	int dataLen = ((int) receiveBuffer[DATALEN]) & 0xff;
	int i;

	log.log(Log.DEBUG, "Read address " + baseAddress + "\n");

	for (i = 0; i < dataLen; i++)
	  codeSpace[baseAddress + i] = receiveBuffer[DATA + i];
  } // recordProgramMemory

  public void readDeviceTransaction() {
	currentTransaction = CommandReadProgramMemory;
	connect();

	lastWriteOffset = device.hiProgramMem;
  } // readDeviceTransaction

  public void readDevice() {
	currentCommand = CommandReadProgramMemory;
	currentOffset = device.loProgramMem;

	percent = 0;
	reportProgress(percent);

	sendCommand(CommandReadProgramMemory, parameters.getMaxPacketSize(), currentOffset, 
				null,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // readDevice

  public void readDeviceNext() {
	currentOffset += parameters.getMaxPacketSize();

	percent = (currentOffset - device.loProgramMem) * 100 /
	  (lastWriteOffset - device.loProgramMem);
	if (percent != lastPercent) {
	  reportProgress(percent);
	  lastPercent = percent;
	}

	if (currentOffset > lastWriteOffset) {
      if (device.hiEeData != device.loEeData)
	    readEeData();
      else
      	readConfig();
	  return;
	}

	sendCommand(CommandReadProgramMemory, parameters.getMaxPacketSize(), 
				currentOffset, null,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // readDeviceNext

  private void eraseDevice() {
	currentCommand = CommandEraseProgramMemory;
	currentOffset = device.loProgramMem;

	percent = 0;
	reportProgress(percent);

	sendCommand(CommandEraseProgramMemory, 1, currentOffset, null,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // eraseDevice

  public void eraseDeviceNext() {
	currentOffset += device.eraseBlock;

	if (currentOffset > lastWriteOffset) {
	  // We are done erasing. Start programming.
	  programDevice();
	  return;
	}

	percent = (currentOffset - device.loProgramMem) * 12 /
	  (lastWriteOffset - device.loProgramMem);
	if (percent != lastPercent) {
	  reportProgress(percent);
	  lastPercent = percent;
	}

	sendCommand(CommandEraseProgramMemory, 1, currentOffset, null,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // eraseDeviceNext

  public void programDeviceTransaction() {
	currentTransaction = CommandWriteProgramMemory;
	connect();

	lastWriteOffset = device.hiProgramMem;
	for ( ; (lastWriteOffset >= device.loProgramMem) &&
			(codeSpace[lastWriteOffset] == ((byte) 0xff)); lastWriteOffset--)
	  ;

	log.log(Log.DEBUG, "Last address = " + lastWriteOffset + "\n");
  } // programDeviceTransaction

  private void programDevice() {
	byte data[] = new byte[WRITE_DATA_SIZE];
	int i;

	currentOffset = device.loProgramMem;

	currentCommand = CommandWriteProgramMemory;

	for (i = 0; i < WRITE_DATA_SIZE; i++) {
	  data[i] = codeSpace[currentOffset + i];
	}

	sendCommand(CommandWriteProgramMemory, WRITE_DATA_SIZE, currentOffset, data,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // programDevice

  public void programDeviceNext() {
	byte data[] = new byte[WRITE_DATA_SIZE];
	int i;

	currentOffset += WRITE_DATA_SIZE;

	if (currentOffset > lastWriteOffset) {
	  if (parameters.getProgramEeData() && device.hiEeData != device.loEeData) {
		setProgressStatus(BootloaderResourceBundle.getString("WRITING_DEVICE_EEPROM_DATA_TEXT"));
		writeEeData(false);
	  }
	  else if (parameters.getProgramConfig()) {
		setProgressStatus(BootloaderResourceBundle.getString("WRITING_DEVICE_CONFIG_TEXT"));
		writeConfig();
	  }
	  else {
		// We are done
		if (parameters.getProgramReset())
		  resetDevice();
		terminateCommand(COMMAND_TERM_REASON_SUCCESS);
	  }
	  return;
	}

	percent = 12 + ((currentOffset - device.loProgramMem) * 88 /
	  (lastWriteOffset - device.loProgramMem));
	if (percent != lastPercent) {
	  reportProgress(percent);
	  lastPercent = percent;
	}

	for (i = 0; i < WRITE_DATA_SIZE; i++) {
	  data[i] = codeSpace[currentOffset + i];
	}

	sendCommand(CommandWriteProgramMemory, WRITE_DATA_SIZE, currentOffset, data,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // programDeviceNext

  public boolean verifyBlock(byte[] receiveBuffer) {
	int i;
	int dataLen = ((int) receiveBuffer[DATALEN]) & 0xff;

	switch(currentCommand) {
	case CommandReadProgramMemory:
	  for (i = 0; i < dataLen; i++) {
		if (receiveBuffer[DATA + i] != codeSpace[currentOffset + i]) {
		  hideProgressDialog();
		  if (cliFilename != null) {
		  	System.err.println(BootloaderResourceBundle.getString("VERIFICATION_FAILED_TEXT") + ": " + (currentOffset + i));
		  	return false;
	      }
		  
		  AlertDialog ad = new AlertDialog(this, BootloaderResourceBundle.getString("VERIFYING_DEVICE_TEXT"),
		                                   BootloaderResourceBundle.getString("VERIFICATION_FAILED_TEXT"),
										   "" + (currentOffset + i),
			                               BootloaderResourceBundle.getString("READ_TEXT") + ": " +
										   Utils.hexToAscii(receiveBuffer[DATA + i]) +
										   ", " + BootloaderResourceBundle.getString("EXPECTED_TEXT") + ": " +
										   Utils.hexToAscii(codeSpace[currentOffset +
																	  i]) + 
										   ".");
		  return false;
		}
	  }
	  break;
	case CommandReadEeData:
	  break;
	case CommandReadConfig:
	  break;
	}

	return true;
  } // verifyBlock

  public void verifyDeviceTransaction() {
	verifyInd = true;
	readDeviceTransaction();
  } // verifyDeviceTransaction

  public void verifyDevice() {
	for (lastWriteOffset = device.hiProgramMem;
		 (lastWriteOffset >= device.loProgramMem) &&
		  (codeSpace[lastWriteOffset] == ((byte) 0xff)); lastWriteOffset--)
	  ;

	log.log(Log.DEBUG, "Last address to verify is " + lastWriteOffset + "\n");

	if (lastWriteOffset < device.loProgramMem) {
	  // Nothing to verify in program memory
	  readEeData();
	  return;
	}

	readDevice();
  } // verifyDevice

  public void recordEeData(byte[] receiveBuffer) {
	StringBuffer buffer = new StringBuffer();
	int address = ((((int) receiveBuffer[ADDRU]) & 0xff) << 16) +
	  ((((int) receiveBuffer[ADDRH]) & 0xff) << 8) +
	  (((int) receiveBuffer[ADDRL]) & 0xff);
	int i;
	int dataLen = ((int) receiveBuffer[DATALEN]) & 0xff;

	if (address < device.loEeData || address > device.hiEeData) {
	  log.log(Log.ERROR, "Error: EEPROM data address " + address +
			  " out of range\n");

	  // Retry
	  retryCommand();
	  return;
	}

	for (i = 0; i < dataLen; i++) {
	  eeData[address - device.loEeData + i] = receiveBuffer[DATA + i];
	}

	// Log data
	buffer.append("EEPROM Data " + Utils.hexToAscii((int) receiveBuffer[ADDRU]) +
				  Utils.hexToAscii((int) receiveBuffer[ADDRH]) +
				  Utils.hexToAscii((int) receiveBuffer[ADDRL]) + ": ");
	for (i = 0; i < dataLen; i++) {
	  buffer.append(Utils.hexToAscii((int) receiveBuffer[DATA + i]));
	}
	buffer.append("\n");

	log.log(Log.INFO, buffer.toString());
  } // recordEeData

  public void readEeData() {
	currentCommand = CommandReadEeData;
	currentOffset = device.loEeData;

	setProgressStatus(BootloaderResourceBundle.getString("READING_DEVICE_EEPROM_DATA_TEXT"));

	percent = 0;
	reportProgress(percent);

	sendCommand(CommandReadEeData, parameters.getMaxPacketSize(), currentOffset,
				null, parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // readEeData

 public void readEeDataNext() {
	currentOffset += parameters.getMaxPacketSize();

	percent = (currentOffset - device.loEeData) * 100 /
	  (device.hiEeData - device.loEeData);
	if (percent != lastPercent) {
	  reportProgress(percent);
	  lastPercent = percent;
	}

	if (currentOffset > device.hiEeData) {
	  readConfig();
	  return;
	}

	sendCommand(CommandReadEeData, parameters.getMaxPacketSize(), currentOffset, 
				null, parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // readEeDataNext

  public void writeEeDataTransaction() {
	currentTransaction = CommandWriteEeData;
	currentOffset = device.loEeData;
	eeDataInd = true;
	connect();
  } // writeEeDataTransaction

  public void writeEeData(boolean ind) {
	byte data[] = new byte[device.writeBlock];
	int i;

	eeDataInd = ind;

	currentCommand = CommandWriteEeData;
	currentOffset = device.loEeData;

	percent = 0;
	reportProgress(percent);

	for (i = 0; i < device.writeBlock; i++) {
	  data[i] = eeData[currentOffset - device.loEeData + i];
	}

	sendCommand(CommandWriteEeData, device.writeBlock, 
				currentOffset, data, 
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // writeEeData

  public void writeEeDataNext() {
	byte data[] = new byte[device.writeBlock];
	int i;

	currentOffset += device.writeBlock;

	percent = (currentOffset - device.loEeData) * 100 /
	  (device.hiEeData - device.loEeData);
	if (percent != lastPercent) {
	  reportProgress(percent);
	  lastPercent = percent;
	}

	if (currentOffset > device.hiEeData) {
	  if (parameters.getProgramConfig() && !eeDataInd) {
		setProgressStatus(BootloaderResourceBundle.getString("WRITING_DEVICE_CONFIG_TEXT"));
		writeConfig();
	  }
	  else {
	    // We are done
	    if (!eeDataInd && parameters.getProgramReset())
		  resetDevice();
		terminateCommand(COMMAND_TERM_REASON_SUCCESS);
	  }
	  return;
	}

	for (i = 0; i < device.writeBlock; i++) {
	  data[i] = eeData[currentOffset - device.loEeData + i];
	}

	sendCommand(CommandWriteEeData, device.writeBlock, currentOffset,
				data,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // writeEeDataNext

  public void runDeviceTransaction() {
	currentTransaction = CommandReset;
	connect();
  } // runDeviceTransaction

  private void runDevice() {
  	if (parameters.getBootMode() == BootloaderParameters.BOOT_MODE_USER) {
  	  writeStartupDelay(0);
  	  return;
    }
  	
    resetDevice();

	terminateCommand(COMMAND_TERM_REASON_SUCCESS);
  } // runDevice

  private void resetDevice() {
	currentCommand = CommandReset;

	sendCommand(CommandReset, 1, 0, null, -1, 0);
  } // resetDevice

  public void writeStartupDelay(int delay) {
  	// Write specified startup delay on last byte of EEPROM.
	byte data[] = new byte[1];

	startupDelay = true;

	currentCommand = CommandWriteEeData;
	currentOffset = device.hiEeData;
	
    data[0] = (byte) delay;

	sendCommand(CommandWriteEeData, 1, 
				currentOffset, data, 
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // writeStartupDelay
	
  public void recordConfig(byte[] receiveBuffer) {
	StringBuffer buffer = new StringBuffer();
	int address = (receiveBuffer[ADDRU] << 16) + (receiveBuffer[ADDRH] << 8) +
	  receiveBuffer[ADDRL];
 
	if (address < device.loConfigMem || address > device.hiConfigMem) {
	  log.log(Log.ERROR, "Error: Config address " + address + " out of range\n");

	  // Retry
	  retryCommand();
	  return;
	}

	config[address - device.loConfigMem] = receiveBuffer[DATA];

	buffer.append("Config " + Utils.hexToAscii((int) receiveBuffer[ADDRU]) +
				  Utils.hexToAscii((int) receiveBuffer[ADDRH]) +
				  Utils.hexToAscii((int) receiveBuffer[ADDRL]));
	buffer.append(": " + Utils.hexToAscii((int) receiveBuffer[DATA]) + "\n");

	log.log(Log.INFO, buffer.toString());
  } // recordConfig

  public void readConfig() {
	currentCommand = CommandReadConfig;
	currentOffset = device.loConfigMem;

	setProgressStatus(BootloaderResourceBundle.getString("READING_DEVICE_CONFIG_TEXT"));

	percent = 0;
	reportProgress(percent);

	sendCommand(CommandReadConfig, 1, currentOffset, null,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // readConfig

  public void readConfigNext() {
	currentOffset++;

	percent = (currentOffset - device.loConfigMem) * 100 /
	  (device.hiConfigMem - device.loConfigMem);
	if (percent != lastPercent) {
	  reportProgress(percent);
	  lastPercent = percent;
	}

	if (currentOffset > device.hiConfigMem) {
	  hideProgressDialog();

	  if (!verifyInd)
		refreshMemory();
	  else {
	  	if (cliFilename != null) {
	      System.out.println(BootloaderResourceBundle.getString("VERIFICATION_SUCCESSFUL_TEXT"));
	  	}
	  	else {
		  AlertDialog ad = new AlertDialog(this, BootloaderResourceBundle.getString("VERIFYING_DEVICE_TEXT"),
										   BootloaderResourceBundle.getString("VERIFICATION_SUCCESSFUL_TEXT"),
										   "",
										   "");
	  	}
		verifyInd = false;
	  }

	  terminateCommand(COMMAND_TERM_REASON_SUCCESS);

	  return;
	}

	sendCommand(CommandReadConfig, 1, currentOffset, null,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // readConfigNext

  public void writeConfigTransaction() {
	currentTransaction = CommandWriteConfig;
	configInd = true;
	connect();
  } // writeConfigTransaction

  private void writeConfig() {
	byte data[] = new byte[1];

	currentCommand = CommandWriteConfig;
	currentOffset = device.loConfigMem;

	percent = 0;
	reportProgress(percent);

	data[0] = config[currentOffset - device.loConfigMem];

	sendCommand(CommandWriteConfig, 1, currentOffset, data,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // writeConfig

  public void writeConfigNext() {
	byte data[] = new byte[1];

	currentOffset++;

	percent = (currentOffset - device.loConfigMem) * 100 /
	  (device.hiConfigMem - device.loConfigMem);
	if (percent != lastPercent) {
	  reportProgress(percent);
	  lastPercent = percent;
	}

	if (currentOffset > device.hiConfigMem) {
	  // We are done
      if (!configInd && parameters.getProgramReset())
		resetDevice();
	  terminateCommand(COMMAND_TERM_REASON_SUCCESS);
	  return;
	}

	data[0] = config[currentOffset - device.loConfigMem];

	sendCommand(CommandWriteConfig, 1, currentOffset, data,
				parameters.getMaxTimeouts(), parameters.getTimeoutDelay());
  } // writeConfigNext

  public String getLastCommand() {
	StringBuffer buffer = new StringBuffer();
	int i;

	for (i = 0; i < lastCommandSentLength; i++) {
	  buffer.append(Utils.hexToAscii((int) lastCommandSent[i]));
	}

	return buffer.toString();
  } // getLastCommand

  public void eraseMemory() {
	int i;

	for (i = device.loProgramMem; i <= device.hiProgramMem; i++) {
	  codeSpace[i] = (byte) 0xff;
	}

	for (i = device.loConfigMem; i <= device.hiConfigMem; i++) {
	  config[i - device.loConfigMem] = (byte) 0xff;
	}

	for (i = device.loEeData; i <= device.hiEeData; i++) {
	  eeData[i - device.loEeData] = (byte) 0xff;
	}
  } // eraseMemory

  public void resizeMemory(int lastAddress) {
	// If device memory space is too small, increase it
	if (device.hiProgramMem < lastAddress)
	  device.hiProgramMem = lastAddress;
  } // resizeMemory

  public void refreshMemory() {
	int i;
	StringBuffer memoryDump = new StringBuffer();

	memoryArea.setText("");
    
	for (i = device.loProgramMem; i <= device.hiProgramMem; i++) {
	  if (i % 16 == 0) {
		if (device.hiProgramMem > 0xffff)
		  memoryDump.append(Utils.hexToAscii(i >> 16) + Utils.hexToAscii(i >> 8) + Utils.hexToAscii(i & 0xff) +
							": ");
		else
		  memoryDump.append(Utils.hexToAscii(i >> 8) + Utils.hexToAscii(i & 0xff) +
							": ");
	  }

	  memoryDump.append(Utils.hexToAscii(codeSpace[i]));

	  if ((i + 1) % 2 == 0)
		memoryDump.append(" ");

	  if ((i + 1) % 16 == 0)
		memoryDump.append("\n");
	}

	memoryDump.append("\nConfig:\n");
	for (i = device.loConfigMem; i <= device.hiConfigMem; i++) {
	  memoryDump.append(Utils.hexToAscii((i >> 16) & 0xff) +
						Utils.hexToAscii((i >> 8) & 0xff) +
						Utils.hexToAscii(i & 0xff) + ": ");

	  memoryDump.append(Utils.hexToAscii(config[i - device.loConfigMem]));

	  memoryDump.append("\n");
	}

	memoryDump.append("\nDevice ID:\n");
	for (i = 0; i <= 1; i++) {
	  memoryDump.append(Utils.hexToAscii(((PicDevice.DEVICE_ID_BASE_ADDRESS + i) >> 16) & 0xff) +
						Utils.hexToAscii(((PicDevice.DEVICE_ID_BASE_ADDRESS + i) >> 8) & 0xff) +
						Utils.hexToAscii((PicDevice.DEVICE_ID_BASE_ADDRESS + i) & 0xff) + ": ");

	  memoryDump.append(Utils.hexToAscii(deviceIdRaw[i]));

	  memoryDump.append("\n");
	}


	memoryDump.append("\nEEPROM Data:\n");
	for (i = device.loEeData; i <= device.hiEeData; i++) {
	  if (i % 8 == 0) {
		memoryDump.append(Utils.hexToAscii(i >> 8) +
						  Utils.hexToAscii(i & 0xff) + ": ");
	  }

	  memoryDump.append(Utils.hexToAscii(eeData[i - device.loEeData]));

	  if ((i + 1) % 2 == 0)
		memoryDump.append(" ");

	  if ((i + 1) % 8 == 0)
		memoryDump.append("\n");
	}

	memoryArea.setText(memoryDump.toString());
  } // refreshMemory

  public byte getProgramMem(int address) {
	if (address > device.hiProgramMem) {
	  log.log(Log.ERROR, "Invalid program address access: " + address + "\n");
	  return 0;
	}

	return codeSpace[address];
  } // getProgramMem

  public void loadProgramMem(int address, byte data) {
	if (address < device.loProgramMem || address > device.hiProgramMem) {
	  log.log(Log.ERROR, "Invalid address access: " + address + "\n");
	  return;
	}

	codeSpace[address] = data;
  } // loadProgramMem

  public byte getConfig(int address) {
	if (address < device.loConfigMem || address > device.hiConfigMem) {
	  log.log(Log.ERROR, "Invalid config address access: " + address + "\n");
	  return 0;
	}

	return config[address - device.loConfigMem];
  } // getConfig

  public void loadConfig(int address, byte data) {
	if (address < device.loConfigMem || address > device.hiConfigMem) {
	  log.log(Log.ERROR, "Invalid config address access: " + address + "\n");
	  return;
	}

	config[address - device.loConfigMem] = data;
  } // loadConfig

  public byte getEeData(int address) {
	if (address < device.loEeData || address > device.hiEeData) {
	  log.log(Log.ERROR, "Invalid EEPROM Data address access: " + address + "\n");
	  return 0;
	}

	return eeData[address - device.loEeData];
  } // getEeData

  public void loadData(int address, byte data) {
	if (address < device.loEeData || address > device.hiEeData) {
	  log.log(Log.ERROR, "Invalid EEPROM data address access: " + address + "\n");
	  return;
	}

	eeData[address - device.loEeData] = data;
  } // loadData

  public PicDevice getDevice() {
	return device;
  } // getDevice

  public String getOperationString() {
	switch(currentCommand) {
	case CommandReadVersion:
	  return new String("Read Firmware Version");
	case CommandReadProgramMemory:
	  if (currentOffset == 0x3fffffe)
		return new String("Read Device ID");
	  return new String("Read Program Memory");
	case CommandWriteProgramMemory:
	  return new String("Write Program Memory");
	case CommandReadEeData:
	  return new String("Read EEPROM Data");
	case CommandWriteEeData:
	  return new String("Write EEPROM Data");
	case CommandReadConfig:
	  return new String("Read Config");
	case CommandWriteConfig:
	  return new String("Write Config");
	}

	return new String("Unknown command");
  } // getOperationString

  public void terminateCommand(int reason) {
	retryTimer.stop();
	currentTransaction = CommandNull;
	currentCommand = CommandNull;
	verifyInd = false;
	eeDataInd = false;
	configInd = false;
	hideProgressDialog();
	reportProgress(0);
	lastPercent = 0;
	connection.scheduleCloseConnection();
	commandTerminated(reason);
	if (isCli()) {
	  if (reason == COMMAND_TERM_REASON_SUCCESS)
	    System.exit(EXIT_NORMAL);
	  else
	  	System.exit(EXIT_ERROR);
	}
  } // terminateCommand

  public int getProgramLastAddress() {
	int lastAddress;

	lastAddress = device.hiProgramMem;
	for ( ; (lastAddress >= device.loProgramMem) &&
			(codeSpace[lastAddress] == ((byte) 0xff)); lastAddress--)
	  ;

	return lastAddress;
  } // getLastAddress

  public boolean isEeDataPresent() {
	int i;

	for (i = device.loEeData; i <= device.hiEeData; i++)
	  if (eeData[i - device.loEeData] != 0xff)
		return true;

	return false;
  } // isEeDataPresent

  public boolean isConfigPresent() {
	int i;

	for (i = device.loConfigMem; i <= device.hiConfigMem; i++)
	  if (config[i - device.loConfigMem] != 0xff)
		return true;

	return false;
  } // isConfigPresent

  public void openConnection() {
	firmwareVersion = 0;
	deviceId = 0;
  } // openConnection
  
  public void closeConnection() {
	if (currentTransaction != CommandNull) {
	  currentTransaction = CommandNull;
	  currentCommand = CommandNull;
	  hideProgressDialog();
	  commandTerminated(COMMAND_TERM_REASON_RESOURCE_NOT_AVAILABLE);
	}
  } // closeConnection
}
