/* @(#)BootloaderParameters.java	0.1 02/10/03 SMI
 *
 * Copyright (c) 2002 Martin Dubuc All Rights Reserved.
 *
 * Martin Dubuc grants you ("Licensee") a non-exclusive, royalty free, license
 * to use, modify and redistribute this software in source and binary
 * code form, provided that i) this copyright notice and license appear
 * on all copies of the software; and ii) Licensee does not utilize the
 * software in a manner which is disparaging to Martin Dubuc.
 *
 * This software is provided "AS IS," without a warranty of any kind.
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. MARTIN DUBUC AND
 * ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THE
 * SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL MARTIN DUBUC OR ITS LICENSORS
 * BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES,
 * HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING
 * OUT OF THE USE OF OR INABILITY TO USE SOFTWARE, EVEN IF MARTIN DUBUC HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control
 * of aircraft, air traffic, aircraft navigation or aircraft
 * communications; or in the design, construction, operation or
 * maintenance of any nuclear facility. Licensee represents and
 * warrants that it will not use or redistribute the Software for such
 * purposes.
 */

import javax.comm.*;

/**
A class that stores bootloader parameters. 
*/
public class BootloaderParameters {
  static final public int LANGUAGE_ENGLISH = 0;
  static final public int LANGUAGE_FRENCH = 1;
  static final public int LANGUAGE_DEFAULT = LANGUAGE_ENGLISH;
	
  static final public int BOOT_MODE_DELAYED = 1;
  static final public int BOOT_MODE_USER = 2;

  static final private int DEFAULT_TIMEOUT_DELAY = 1000;
  static final private int DEFAULT_MAX_TIMEOUTS = 3;
  static final private int DEFAULT_MAX_PACKET_SIZE = 128;
  static final private int DEFAULT_LOG_LEVEL = Log.NONE;
  static final private int DEFAULT_BOOT_MODE = BOOT_MODE_DELAYED;

  static final int NUM_RECENT_PROGRAM = 5;

  private int language;

  private String portName;
  private int baudRate;
  private int flowControlIn;
  private int flowControlOut;
  private int databits;
  private int stopbits;
  private int parity;

  private int timeoutDelay;
  private int maxTimeouts;
  private int maxPacketSize;

  private int logLevel;

  private String programDir;

  private String programName;

  private String recentProgramDir[];
  private String recentProgramName[];

  private int bootMode;

  private boolean programAutoload;

  private boolean programReset;

  private boolean programEeData;

  private boolean programConfig;

  private boolean fileDialogUseSwing;

  private boolean pic18fxx2Ind;
  
  /**
  Default constructer. Sets parameters to no port, 9600 baud, no flow 
  control, 8 data bits, 1 stop bit, no parity, disable autoload,
  disable autoreset, empty path and filename, use Swing for File 
  Dialog.
  */
  public BootloaderParameters () {
    this(LANGUAGE_DEFAULT,
    "",
    9600,
    SerialPort.FLOWCONTROL_NONE,
    SerialPort.FLOWCONTROL_NONE,
    SerialPort.DATABITS_8,
    SerialPort.STOPBITS_1,
    SerialPort.PARITY_NONE,
    DEFAULT_TIMEOUT_DELAY,
    DEFAULT_MAX_TIMEOUTS,
    DEFAULT_MAX_PACKET_SIZE,
    DEFAULT_LOG_LEVEL,
    DEFAULT_BOOT_MODE,
    false,
    false,
    false,
    false, "", "",
    true,
    true);		   
  } // BooloaderParameters		

  /**
  Paramaterized constructer.

  @param language The language.
  @param portName The name of the port.
  @param baudRate The baud rate.
  @param flowControlIn Type of flow control for receiving.
  @param flowControlOut Type of flow control for sending.
  @param databits The number of data bits.
  @param stopbits The number of stop bits.
  @param parity The type of parity.
  @param timeoutDelay The delay before bootloader times out waiting for a response from
  PIC device.
  @param maxTimeouts The maximum number of timeouts before PIC is declared dead.
  @param maxPacketSize The maximum packet size.
  @param logLevel The logging level.
  @param bootMode The boot mode.
  @param programAutoload Indicates if bootloader should automatically reload program.
  @param programReset Indicates if bootloader should automatically reset PIC after
  loading program.
  @param programConfig Indicates if bootloader should program Config.
  @param programEeData Indicates if bootloader should program EEPROM Data.
  @param programDir Program directory.
  @param programName Program name.
  @param fileDialogUseSwing Indicates whether Swing should be used for
  File Dialog.
  @param pic18fxx2Ind Indicates whether Jolt should display PIC18FXX2 family over
  PIC18FXX39.
  */
  public BootloaderParameters(int language,
                              String portName, 
                              int baudRate,
                              int flowControlIn,
                              int flowControlOut,
                              int databits,
                              int stopbits,
                              int parity,
                              int timeoutDelay,
                              int maxTimeouts,
                              int maxPacketSize,
                              int logLevel,
                              int bootMode,
                              boolean programAutoload,
                              boolean programReset,
                              boolean programConfig,
                              boolean programEeData,
                              String programDir,
                              String programName,
                              boolean fileDialogUseSwing,
                              boolean pic18fxx2Ind) {

    this.language = language;
    this.portName = portName;
    this.baudRate = baudRate;
    this.flowControlIn = flowControlIn;
    this.flowControlOut = flowControlOut;
    this.databits = databits;
    this.stopbits = stopbits;
    this.parity = parity;
    this.timeoutDelay = timeoutDelay;
    this.maxTimeouts = maxTimeouts;
    this.maxPacketSize = maxPacketSize;
    this.logLevel = logLevel;
    this.bootMode = bootMode;
    this.programAutoload = programAutoload;
    this.programReset = programReset;
    this.programConfig = programConfig;
    this.programEeData = programEeData;
    this.programDir = programDir;
    this.programName = programName;
    this.fileDialogUseSwing = fileDialogUseSwing;
    this.pic18fxx2Ind = pic18fxx2Ind;
    
    this.recentProgramDir = new String[NUM_RECENT_PROGRAM];
    this.recentProgramName = new String[NUM_RECENT_PROGRAM];
  } // BootloaderParameters

  /**
   Sets language.
   @param language New language.
   */
  public void setLanguage(int language) {
	this.language = language;
  } // setLanguage

  /**
   Sets language.
   @param language New language.
   */
  public void setLanguage(String language) {
    this.language = Integer.parseInt(language);
  } // setLanguage

  /**
   Gets language as an <code>int</code>.
   @return Current language.
   */
  public int getLanguage() {
    return language;
  } // getLanguage

  /**
   Gets language as a <code>String</code>.
   @return Current language.
   */
  public String getLanguageString() {
    return Integer.toString(language);
  } // getLanguageString

  /**
  Sets port name.
  @param portName New port name.
  */
  public void setPortName(String portName) {
    this.portName = portName;
  } // setPortName

  /**
  Gets port name.
  @return Current port name.
  */
  public String getPortName() {
    return portName;
  } // getPortName

  /**
  Sets baud rate.
  @param baudRate New baud rate.
  */
  public void setBaudRate(int baudRate) {
    this.baudRate = baudRate;
  } // setBaudRate

  /**
  Sets baud rate.
  @param baudRate New baud rate.
  */
  public void setBaudRate(String baudRate) {
    this.baudRate = Integer.parseInt(baudRate);
  } // setBaudRate

  /**
  Gets baud rate as an <code>int</code>.
  @return Current baud rate.
  */
  public int getBaudRate() {
    return baudRate;
  } // getBaudRate

  /**
  Gets baud rate as a <code>String</code>.
  @return Current baud rate.
  */
  public String getBaudRateString() {
    return Integer.toString(baudRate);
  } // getBaudRateString

  /**
  Sets flow control for reading.
  @param flowControlIn New flow control for reading type.
  */
  public void setFlowControlIn(int flowControlIn) {
    this.flowControlIn = flowControlIn;
  } // setFlowControlIn

  /**
  Sets flow control for reading.
  @param flowControlIn New flow control for reading type.
  */
  public void setFlowControlIn(String flowControlIn) {
    this.flowControlIn = stringToFlow(flowControlIn);
  } // setFlowControlIn

  /** 
  Gets flow control for reading as an <code>int</code>.
  @return Current flow control type.
  */
  public int getFlowControlIn() {
    return flowControlIn;
  } // getFlowControlIn

  /** 
  Gets flow control for reading as a <code>String</code>.
  @return Current flow control type.
  */
  public String getFlowControlInString() {
    return flowToString(flowControlIn);
  } // getFlowControlInString

  /**
  Sets flow control for writing.
  @param flowControlIn New flow control for writing type.
  */
  public void setFlowControlOut(int flowControlOut) {
    this.flowControlOut = flowControlOut;
  }

  /**
  Sets flow control for writing.
  @param flowControlIn New flow control for writing type.
  */
  public void setFlowControlOut(String flowControlOut) {
    this.flowControlOut = stringToFlow(flowControlOut);
  }

  /** 
  Gets flow control for writing as an <code>int</code>.
  @return Current flow control type.
  */
  public int getFlowControlOut() {
    return flowControlOut;
  }

  /** 
  Gets flow control for writing as a <code>String</code>.
  @return Current flow control type.
  */
  public String getFlowControlOutString() {
    return flowToString(flowControlOut);
  }

  /** 
  Sets data bits.
  @param databits New data bits setting.
  */
  public void setDatabits(int databits) {
    this.databits = databits;
  }

  /** 
  Sets data bits.
  @param databits New data bits setting.
  */
  public void setDatabits(String databits) {
    if (databits.equals("5")) {
      this.databits = SerialPort.DATABITS_5;
    }
    if (databits.equals("6")) {
      this.databits = SerialPort.DATABITS_6;
    }
    if (databits.equals("7")) {
      this.databits = SerialPort.DATABITS_7;
    }
    if (databits.equals("8")) {
      this.databits = SerialPort.DATABITS_8;
    }
  }

  /**
  Gets data bits as an <code>int</code>.
  @return Current data bits setting.
  */
  public int getDatabits() {
    return databits;
  }

  /**
  Gets data bits as a <code>String</code>.
  @return Current data bits setting.
  */
  public String getDatabitsString() {
    switch(databits) {
    case SerialPort.DATABITS_5:
      return "5";
    case SerialPort.DATABITS_6:
      return "6";
    case SerialPort.DATABITS_7:
      return "7";
    case SerialPort.DATABITS_8:
      return "8";
    default:
      return "8";
    }
  }

    /**
    Sets stop bits.
    @param stopbits New stop bits setting.
    */
    public void setStopbits(int stopbits) {
	this.stopbits = stopbits;
    }

    /**
    Sets stop bits.
    @param stopbits New stop bits setting.
    */
    public void setStopbits(String stopbits) {
	if (stopbits.equals("1")) {
	    this.stopbits = SerialPort.STOPBITS_1;
	}
	if (stopbits.equals("1.5")) {
	    this.stopbits = SerialPort.STOPBITS_1_5;
	}
	if (stopbits.equals("2")) {
	    this.stopbits = SerialPort.STOPBITS_2;
	}
    }

    /**
    Gets stop bits setting as an <code>int</code>.
    @return Current stop bits setting.
    */
    public int getStopbits() {
	return stopbits;
    }

    /**
    Gets stop bits setting as a <code>String</code>.
    @return Current stop bits setting.
    */
    public String getStopbitsString() {
	switch(stopbits) {
	    case SerialPort.STOPBITS_1:
		return "1";
	    case SerialPort.STOPBITS_1_5:
		return "1.5";
	    case SerialPort.STOPBITS_2:
		return "2";
	    default:
		return "1";
	}
    }

    /**
    Sets parity setting.
    @param parity New parity setting.
    */
    public void setParity(int parity) {
	this.parity = parity;
    }

    /**
    Sets parity setting.
    @param parity New parity setting.
    */
    public void setParity(String parity) {
	if (parity.equals("None")) {
	    this.parity = SerialPort.PARITY_NONE;
	}
	if (parity.equals("Even")) {
	    this.parity = SerialPort.PARITY_EVEN;
	}
	if (parity.equals("Odd")) {
	    this.parity = SerialPort.PARITY_ODD;
	}
    }

    /**
    Gets parity setting as an <code>int</code>.
    @return Current parity setting.
    */
    public int getParity() {
	return parity;
    }

    /**
    Gets parity setting as a <code>String</code>.
    @return Current parity setting.
    */
    public String getParityString() {
	switch(parity) {
	    case SerialPort.PARITY_NONE:
		return "None";
 	    case SerialPort.PARITY_EVEN:
		return "Even";
	    case SerialPort.PARITY_ODD:
		return "Odd";
	    default:
		return "None";
	}
    }

    /**
    Converts a <code>String</code> describing a flow control type to an
    <code>int</code> type defined in <code>SerialPort</code>.
    @param flowControl A <code>string</code> describing a flow control type.
    @return An <code>int</code> describing a flow control type.
    */
    private int stringToFlow(String flowControl) {
	if (flowControl.equals("None")) {
	    return SerialPort.FLOWCONTROL_NONE;
	}
	if (flowControl.equals("Xon/Xoff Out")) {
	    return SerialPort.FLOWCONTROL_XONXOFF_OUT;
	}
	if (flowControl.equals("Xon/Xoff In")) {
	    return SerialPort.FLOWCONTROL_XONXOFF_IN;
	}
	if (flowControl.equals("RTS/CTS In")) {
	    return SerialPort.FLOWCONTROL_RTSCTS_IN;
	}
	if (flowControl.equals("RTS/CTS Out")) {
	    return SerialPort.FLOWCONTROL_RTSCTS_OUT;
	}
	return SerialPort.FLOWCONTROL_NONE;
    }

    /**
    Converts an <code>int</code> describing a flow control type to a 
    <code>String</code> describing a flow control type.
    @param flowControl An <code>int</code> describing a flow control type.
    @return A <code>String</code> describing a flow control type.
    */
    String flowToString(int flowControl) {
	switch(flowControl) {
	    case SerialPort.FLOWCONTROL_NONE:
		return "None";
	    case SerialPort.FLOWCONTROL_XONXOFF_OUT:
		return "Xon/Xoff Out";
	    case SerialPort.FLOWCONTROL_XONXOFF_IN:
		return "Xon/Xoff In";
	    case SerialPort.FLOWCONTROL_RTSCTS_IN:
		return "RTS/CTS In";
	    case SerialPort.FLOWCONTROL_RTSCTS_OUT:
		return "RTS/CTS Out";
	    default:
		return "None";
	}
    }

  /**
  Sets timeout delay.
  @param timeoutDelay New timeout delay.
  */
  public void setTimeoutDelay(String timeoutDelay) {
    this.timeoutDelay = Integer.parseInt(timeoutDelay);
  } // setTimeoutDelay

  /**
  Gets timeout delay as an <code>int</code>.
  @return Current timeout delay.
  */
  public int getTimeoutDelay() {
    return timeoutDelay;
  } // getTimeoutDelay

  /**
  Gets timeout delay as a <code>String</code>.
  @return Current timeout delay.
  */
  public String getTimeoutDelayString() {
    return Integer.toString(timeoutDelay);
  } // getTimeoutDelayString

  /**
  Sets maximum number of timeouts.
  @param timeoutDelay New maximum number of timeouts.
  */
  public void setMaxTimeouts(String maxTimeouts) {
    this.maxTimeouts = Integer.parseInt(maxTimeouts);
  } // setMaxTimeouts

  /**
  Gets maximum number of timeouts as an <code>int</code>.
  @return Current maximum number of timeouts.
  */
  public int getMaxTimeouts() {
    return maxTimeouts;
  } // getMaxTimeouts

  /**
  Gets maximum number of timeouts as a <code>String</code>.
  @return Current maximum number of timeouts.
  */
  public String getMaxTimeoutsString() {
    return Integer.toString(maxTimeouts);
  } // getMaxTimeoutsString

  /**
  Sets maximum packet size.
  @param maxPacketSize New maximum packet size.
  */
  public void setMaxPacketSize(String maxPacketSize) {
    this.maxPacketSize = Integer.parseInt(maxPacketSize);
  } // setMaxPacketSize

  /**
  Gets maximum packet size as an <code>int</code>.
  @return Current maximum packet size.
  */
  public int getMaxPacketSize() {
    return maxPacketSize;
  } // getMaxPacketSize

  /**
  Gets maximum packet size as a <code>String</code>.
  @return Current maximum packet size.
  */
  public String getMaxPacketSizeString() {
    return Integer.toString(maxPacketSize);
  } // getMaxPacketString

  /** 
  Sets log level.
  @param logLevel New log level.
  */
  public void setLogLevel(String logLevel) {
    this.logLevel = Integer.parseInt(logLevel);
  } // setLogLevel

  /**
  Gets log level as an <code>int</code>.
  @return Current log level.
  */
  public int getLogLevel() {
    return logLevel;
  } // getLogLevel

  /**
  Gets log level as a <code>String</code>.
  @return Current log level.
  */
  public String getLogLevelString() {
    return Integer.toString(logLevel);
  } // getLogLevelString

  /** 
  Sets program autoload indicator.
  @param programAutoload New program autoload indicator.
  */
  public void setProgramAutoload(String programAutoload) {
    this.programAutoload = Boolean.valueOf(programAutoload).booleanValue();
  } // setProgramAutoload

  /**
  Gets program autoload indicator as an <code>int</code>.
  @return Current program autoload.
  */
  public boolean getProgramAutoload() {
    return programAutoload;
  } // getProgramAutoload

  /**
  Gets program autoload indicator as a <code>String</code>.
  @return Current program autoload indicator.
  */
  public String getProgramAutoloadString() {
    return Boolean.toString(programAutoload);
  } // getProgramAutoloadString

  /** 
  Sets program reset indicator.
  @param programReset New program reset indicator.
  */
  public void setProgramReset(String programReset) {
    this.programReset = Boolean.valueOf(programReset).booleanValue();
  } // setProgramReset

  /**
  Gets program reset indicator as an <code>int</code>.
  @return Current program reset.
  */
  public boolean getProgramReset() {
    return programReset;
  } // getProgramReset

  /**
  Gets program reset indicator as a <code>String</code>.
  @return Current program reset indicator.
  */
  public String getProgramResetString() {
    return Boolean.toString(programReset);
  } // getProgramResetString

  /** 
  Sets program Config indicator.
  @param programConfig New program Config indicator.
  */
  public void setProgramConfig(String programConfig) {
    this.programConfig = Boolean.valueOf(programConfig).booleanValue();
  } // setProgramConfig

  /**
  Gets program Config indicator as an <code>int</code>.
  @return Current program Config.
  */
  public boolean getProgramConfig() {
    return programConfig;
  } // getProgramConfig

  /**
  Gets program Config indicator as a <code>String</code>.
  @return Current program Config indicator.
  */
  public String getProgramConfigString() {
    return Boolean.toString(programConfig);
  } // getProgramConfigString

  /** 
  Sets program EE Data indicator.
  @param programEeData New program EE Data indicator.
  */
  public void setProgramEeData(String programEeData) {
    this.programEeData = Boolean.valueOf(programEeData).booleanValue();
  } // setProgramEeData

  /**
  Gets program EE Data indicator as an <code>int</code>.
  @return Current program EE Data.
  */
  public boolean getProgramEeData() {
    return programEeData;
  } // getProgramEeData

  /**
  Gets program EE Data indicator as a <code>String</code>.
  @return Current program EE Data indicator.
  */
  public String getProgramEeDataString() {
    return Boolean.toString(programEeData);
  } // getProgramEeDataString

  /** 
  Sets boot mode.
  @param bootMode New boot mode.
  */
  public void setBootMode(int bootMode) {
    this.bootMode = bootMode;
  } // setBootMode

  /** 
  Sets boot mode.
  @param bootMode New boot mode.
  */
  public void setBootMode(String bootMode) {
    this.bootMode = Integer.parseInt(bootMode);
  } // setBootMode

  /**
  Gets bootMode as an <code>int</code>.
  @return Current boot mode.
  */
  public int getBootMode() {
    return bootMode;
  } // getBootMode

  /**
  Gets boot mode as a <code>String</code>.
  @return Current boot mode.
  */
  public String getBootModeString() {
    return Integer.toString(bootMode);
  } // getBootModeString

  /**
  Sets program directory.
  @param programDir New program directory.
  */
  public void setProgramDir(String programDir) {
    this.programDir = programDir;
  } // setProgramDir

  /**
  Gets program directory.
  @return Current program directory.
  */
  public String getProgramDir() {
    return programDir;
  } // getProgramDir

  /**
  Sets program name.
  @param programName New program name.
  */
  public void setProgramName(String programName) {
    this.programName = programName;
  } // setProgramName

  /**
  Gets program name.
  @return Current program name.
  */
  public String getProgramName() {
    return programName;
  } // getProgramName

  /**
  Sets file dialog use Swing parameter.
  @param fileDialogUseSwing New file dialog use Swing parameter.
  */
  public void setFileDialogUseSwing(String fileDialogUseSwing) {
    this.fileDialogUseSwing = Boolean.valueOf(fileDialogUseSwing).booleanValue();
  } // setOpenUseSwing

  /**
  Gets file dialog use swing parameter as a <code>boolean</code>.
  @return Current file dialog use Swing parameter.
  */
  public boolean getFileDialogUseSwing() {
    return fileDialogUseSwing;
  } // getFileDialogUseSwing

  /**
  Gets file dialog use swing parameter as a <code>String</code>.
  @return Current file dialog use Swing parameter.
  */
  public String getFileDialogUseSwingString() {
    return Boolean.toString(fileDialogUseSwing);
  } // getFileDialogUseSwingString

  /**
  Sets recent program directory and filename.
  @param progDir New recent program directory parameter.
  @param progName New recent program filename parameter.
  */
  public void setRecentProgram(String progDir, String progName) {
    int i;
    String nextProgDir, nextProgName;
    String dir, name;

    nextProgDir = this.recentProgramDir[0];
    nextProgName = this.recentProgramName[0];

    for (i = 0; i < NUM_RECENT_PROGRAM - 1; i++) {
      if (nextProgDir == null)
        break;

      if (nextProgDir.equals(progDir) &&
          nextProgName.equals(progName))
        break;

      dir = this.recentProgramDir[i + 1];
      name = this.recentProgramName[i + 1];

      this.recentProgramDir[i + 1] = nextProgDir;
      this.recentProgramName[i + 1] = nextProgName;

      nextProgDir = dir;
      nextProgName = name;
    }

    this.recentProgramDir[0] = progDir;
    this.recentProgramName[0] = progName;
  } // setRecentProgram

  /**
  Gets recent program directory parameter at specified index as a <code>String</code>.
  @return Recent program directory parameter.
  */
  public String getRecentProgramDir(int index) {
    if ((index < 0) || (index >= NUM_RECENT_PROGRAM))
      return null;

    return recentProgramDir[index];
  } // getRecentProgramDir

  /**
  Gets recent program name parameter at specified index as a <code>String</code>.
  @return Recent program name parameter.
  */
  public String getRecentProgramName(int index) {
    if ((index < 0) || (index >= NUM_RECENT_PROGRAM))
      return null;

    return recentProgramName[index];
  } // getRecentProgramName
  
  /**
   Sets PIC18FXX2 indicator parameter.
   @param pic18fxx2Ind New PIC18FXX2 indicator parameter.
   */
  public void setPic18fxx2Ind(String pic18fxx2Ind) {
  	this.pic18fxx2Ind = Boolean.valueOf(pic18fxx2Ind).booleanValue();
  } // setPic18fxx2Ind

  /**
   Sets PIC18FXX2 indicator parameter.
   @param pic18fxx2Ind New PIC18FXX2 indicator parameter.
   */
  public void setPic18fxx2Ind(boolean pic18fxx2Ind) {
	this.pic18fxx2Ind = pic18fxx2Ind;
  } // setPic18fxx2Ind

  /**
   Gets PIC18FXX2 indicator parameter as a <code>boolean</code>.
   @return Current PIC18FXX2 indicator parameter.
   */
  public boolean getPic18fxx2Ind() {
	return pic18fxx2Ind;
  } // getPic18fxx2Ind

  /**
   Gets PIC18FXX2 indicator parameter as a <code>String</code>.
   @return Current PIC18FXX2 indicator parameter.
   */
  public String getPic18fxx2IndString() {
  	return Boolean.toString(pic18fxx2Ind);
  } // getPic18fxx2IndString
} // BootloaderParameters
