/* @(#)SerialConnection.java	0.1 02/09/26 SMI
 *
 * Copyright (c) 2002, 2003 Martin Dubuc All Rights Reserved.
 *
 * Martin Dubuc grants you ("Licensee") a non-exclusive, royalty free, license
 * to use, modify and redistribute this software in source and binary
 * code form, provided that i) this copyright notice and license appear
 * on all copies of the software; and ii) Licensee does not utilize the
 * software in a manner which is disparaging to Martin Dubuc.
 *
 * This software is provided "AS IS," without a warranty of any kind.
 * ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND WARRANTIES,
 * INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE OR NON-INFRINGEMENT, ARE HEREBY EXCLUDED. MARTIN DUBUC
 * AND ITS LICENSORS SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY
 * LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING THE
 * SOFTWARE OR ITS DERIVATIVES. IN NO EVENT WILL MARTIN DUBUC OR ITS LICENSORS
 * BE LIABLE FOR ANY LOST REVENUE, PROFIT OR DATA, OR FOR DIRECT,
 * INDIRECT, SPECIAL, CONSEQUENTIAL, INCIDENTAL OR PUNITIVE DAMAGES,
 * HOWEVER CAUSED AND REGARDLESS OF THE THEORY OF LIABILITY, ARISING
 * OUT OF THE USE OF OR INABILITY TO USE SOFTWARE, EVEN IF MARTIN DUBUC
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 *
 * This software is not designed or intended for use in on-line control
 * of aircraft, air traffic, aircraft navigation or aircraft
 * communications; or in the design, construction, operation or
 * maintenance of any nuclear facility. Licensee represents and
 * warrants that it will not use or redistribute the Software for such
 * purposes.
 */

import javax.comm.*;
import java.io.*;
import java.awt.event.*;
import java.util.TooManyListenersException;
import javax.swing.*;

/**
A class that handles the details of a serial connection.
Holds the state of the connection.
*/
public class SerialConnection implements SerialPortEventListener, 
					 CommPortOwnershipListener {
  
  public final static int MAX_BUFFER_LENGTH = 256;
  
  public final static int STX = 0x0f;
  public final static int ETX = 0x04;
  public final static int DLE = 0x05;

  private Bootloader parent;
  private BootloaderParameters parameters;
  
  private Timer closeConnectionTimer;

  private OutputStream os;
  private InputStream is;

  private CommPortIdentifier portId;
  private SerialPort sPort;

  private boolean open;

  private Log log;

  private int numStx;
  private boolean dleFlag;
  private int checksum;

  private byte receiveBuffer[];
  private int receiveBufferLength;

  /**
  Creates a SerialConnection object and initializes variables passed in
  as params.

  @param parent A Bootloader object.
  @param parameters A BootloaderParameters object.
  @param memoryArea The JTextArea that contains the memory dump.
  @param log The logging object.
  */
  public SerialConnection(Bootloader parent,
                          BootloaderParameters parameters,
                          Log log) {

    this.parent = parent;
    this.parameters = parameters;
    this.log = log;
    
    open = false;

	receiveBuffer = new byte[MAX_BUFFER_LENGTH];
	receiveBufferLength = 0;
	checksum = 0;
	numStx = 0;
	dleFlag = false;

    ActionListener closeConnectionAction = new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
        closeConnection();
      }
    };

    closeConnectionTimer = new Timer(1, closeConnectionAction);
    closeConnectionTimer.setRepeats(false);
  } // SerialConnection

  /**
  Attempts to open a serial connection and streams using the parameters
  in the BootloaderParameters object. If it is unsuccesfull at any step it
  returns the port to a closed state, throws a 
  <code>SerialConnectionException</code>, and returns.

  Gives a timeout of 30 seconds on the portOpen to allow other applications
  to reliquish the port if have it open and no longer need it.
  */
  public void openConnection() throws PortInUseException, 
    SerialConnectionException {

    // Obtain a CommPortIdentifier object for the port you want to open.
    try {
      portId = 
        CommPortIdentifier.getPortIdentifier(parameters.getPortName());
    }
    catch (NoSuchPortException e) {
      throw new SerialConnectionException(e.getMessage());
    }

    // Open the port represented by the CommPortIdentifier object.
    try {
      sPort = (SerialPort) portId.open(parent.getAppId(), 1000); // 1 second timeout
    }
    catch (PortInUseException e) {
      if (!e.currentOwner.equals(parent.getAppId()))
        throw e;
    }
  
    // Set the parameters of the connection. If they won't set, close the
    // port before throwing an exception.
    try {
      setConnectionParameters();
    }
    catch (SerialConnectionException e) {	
      sPort.close();
      throw e;
    }

    // Open the input and output streams for the connection. If they won't
    // open, close the port before throwing an exception.
    try {
      os = sPort.getOutputStream();
      is = sPort.getInputStream();
    }
    catch (IOException e) {
      sPort.close();
      throw new SerialConnectionException("Error opening I/O streams");
    }

    // Add this object as an event listener for the serial port.
    try {
      sPort.addEventListener(this);
    }
    catch (TooManyListenersException e) {
      sPort.close();
      throw new SerialConnectionException("Too many listeners added");
    }

    // Set notifyOnDataAvailable to true to allow event driven input.
    sPort.notifyOnDataAvailable(true);

    // Set receive timeout to allow breaking out of polling loop during
    // input handling.
    try {
      sPort.enableReceiveTimeout(30);
    }
    catch (UnsupportedCommOperationException e) {
    }

    // Add ownership listener to allow ownership event handling.
    portId.addPortOwnershipListener(this);

    open = true;

    parent.openConnection();
  } // openConnection

  public void scheduleCloseConnection() {
    closeConnectionTimer.start();
  } // scheduleCloseConnection

  /**
  Sets the connection parameters to the setting in the parameters object.
  If set fails return the parameters object to origional settings and
  throw exception.
  */
  public void setConnectionParameters() throws SerialConnectionException {
    // Save state of parameters before trying a set.
    int oldBaudRate = sPort.getBaudRate();
    int oldDatabits = sPort.getDataBits();
    int oldStopbits = sPort.getStopBits();
    int oldParity   = sPort.getParity();
    int oldFlowControl = sPort.getFlowControlMode();

    // Set connection parameters, if set fails return parameters object
    // to original state.
    try {
      sPort.setSerialPortParams(parameters.getBaudRate(),
                                parameters.getDatabits(),
                                parameters.getStopbits(),
                                parameters.getParity());
    }
    catch (UnsupportedCommOperationException e) {
      parameters.setBaudRate(oldBaudRate);
      parameters.setDatabits(oldDatabits);
      parameters.setStopbits(oldStopbits);
      parameters.setParity(oldParity);
      throw new SerialConnectionException("Unsupported parameter");
    }

    // Set flow control.
    try {
      sPort.setFlowControlMode(parameters.getFlowControlIn() |
                               parameters.getFlowControlOut());
    }
    catch (UnsupportedCommOperationException e) {
      throw new SerialConnectionException("Unsupported flow control");
    }
  } // setConnectionParameters

  /**
  Close the port and clean up associated elements.
  */
  public void closeConnection() {
    // If port is alread closed just return.
    if (!open) {
      return;
    }

    parent.closeConnection();
    
    // Check to make sure sPort has reference to avoid a NPE.
    if (sPort != null) {
      try {
        // Close the I/O streams.
        os.close();
        is.close();
      } catch (IOException e) {
        System.err.println(e);
      }

      // Close the port.
      sPort.close();

      // Remove the ownership listener.
      portId.removePortOwnershipListener(this);
    }

    open = false;
  } // closeConnection

  /**
  Send bytes to the serial port.
  */
  public void sendBytes(String buffer) {
    int i;
    int len;
    int hex;
    StringBuffer data;

    if (!open) {
      return;
    }

    len = buffer.length();
    data = new StringBuffer("00");
    for (i = 0; i < len; i += 2) {
      data.setCharAt(0, buffer.charAt(i));
      if (i + 1 != len) {
        data.setCharAt(1, buffer.charAt(i + 1));
      }
      else
        data.delete(1, 1);

      hex = Utils.asciiToHex(data);

      try {
        os.write((int) hex);
      } catch (IOException e) {
        System.err.println("OutputStream write error: " + e);
      }
    }
  }

  /**
  Reports the open status of the port.
  @return true if port is open, false if port is closed.
  */
  public boolean isOpen() {
    return open;
  }

  /**
  Handles SerialPortEvents. The DATA_AVAILABLE SerialPortEvents are the only
  events that this program is registered to listen for. During 
  DATA_AVAILABLE the port buffer is read until it is drained, when no more
  data is available and 30 ms has passed the method returns.
  */
  public void serialEvent(SerialPortEvent e) {
    // Create a StringBuffer and int to receive input data.
    StringBuffer inputBuffer = new StringBuffer();
    int newData = 0;

    // Determine type of event.
    switch (e.getEventType()) {
    // Read data until -1 is returned.
    case SerialPortEvent.DATA_AVAILABLE:
      /* Could use is.available() to figure out how many characters 
         are buffered */
      while (newData != -1) {
        try {
          newData = is.read();
        } catch (IOException ex) {
          System.err.println(ex);
          return;
        }

        if (newData == -1) {
          break;
        }

        // Log data
        if (newData == STX) {
          if (log.getSeverityLevel() != Log.NONE)
            inputBuffer.append("<STX> ");
        } else if (newData == ETX && !dleFlag) {
          if (log.getSeverityLevel() != Log.NONE)
            inputBuffer.append("<ETX>\n");
        } else if (newData == DLE) {
          if (log.getSeverityLevel() != Log.NONE)
            inputBuffer.append("<DLE> ");
        } else {
          if (log.getSeverityLevel() != Log.NONE) {
            inputBuffer.append(Utils.hexToAscii((int) newData));
            inputBuffer.append((char) ' ');
          }
        }

        // Process data
        if (newData == DLE) {
          if (!dleFlag) {
            // Record escape sequence
            dleFlag = true;
            continue;
          }
        }

        if (!dleFlag) {
          if (newData != STX && numStx < 2) {
            // Restart receive packet
            numStx = 0;
            checksum = 0;
            receiveBufferLength = 0;
            continue;
          }

          if (newData == STX) {
            if (numStx < 2) {
              numStx++;
            }
            else {
              // Restart receive packet
              numStx = 1;
              checksum = 0;
              receiveBufferLength = 0;
            }
            continue;
          }

          if (newData == ETX) {
            if (((~checksum + 1) & 0xff) == 0) {
              if (log.getSeverityLevel() != Log.NONE) {
                log.log(Log.DEBUG, inputBuffer.toString());
                inputBuffer.delete(0, inputBuffer.length());
              }
              parent.processPacket(receiveBuffer);
            }
            else {
              if (log.getSeverityLevel() != Log.NONE) {
                log.log(Log.DEBUG, inputBuffer.toString());
                inputBuffer.delete(0, inputBuffer.length() - 1);
              }
              log.log(Log.INFO, "Bad checksum (" + ((~checksum + 1) & 0xff) +
                      "): discard packet\n");
              parent.retryCommand();
            }

            // Restart packet receive
            numStx = 0;
            checksum = 0;
            receiveBufferLength = 0;
            continue;
          }
        }

        dleFlag = false;

        checksum += newData;

        if (receiveBufferLength >= MAX_BUFFER_LENGTH) {
          // Restart packet receive
          log.log(Log.ERROR, "Reached maximum packet size: discarding packet\n");
          numStx = 0;
          checksum = 0;
          receiveBufferLength = 0;
          continue;
        }

        receiveBuffer[receiveBufferLength++] = (byte) newData;
      }
      break;
    }

    if (log.getSeverityLevel() != Log.NONE) {
      if (inputBuffer.length() != 0) {
        // Log received data.
        log.log(Log.DEBUG, inputBuffer.toString());
      }
    }
  } // serialEvent

  public void sendData(int data) {
	try {
	  if (data == STX || data == ETX || data == DLE) {
		// Insert DLE
		os.write((int) DLE);
	  }
	  os.write((int) data);
	}
	catch (IOException e) {
	  System.err.println("OutputStream write error: " + e);
	}
  } // sendData

  public void sendRawByte(int data) {
  	try {
  	  os.write(data);
	}
	catch (IOException e) {
	  System.err.println("OutputStream write error: " + e);
	}
  } // sendRawByte
    
  /**
  Handles ownership events. If a PORT_OWNERSHIP_REQUESTED event is
  received a dialog box is created asking the user if they are 
  willing to give up the port. No action is taken on other types
  of ownership events.
  */
  public void ownershipChange(int type) {
  	if (portId != null && portId.getCurrentOwner().equals(parent.getAppId())) {
  	  // This application is the one requesting the port
      return;
  	}
  	
    if (type == CommPortOwnershipListener.PORT_OWNERSHIP_REQUESTED) {
      PortRequestedDialog prd = new PortRequestedDialog(parent);
    }
  } // ownershipChange
} // SerialConnection
